import re
from enum import Enum

import numpy as np
import pandas as pd

from htimeseries import HTimeseries

methods = {
    "sum": pd.Series.sum,
    "mean": pd.Series.mean,
    "max": pd.Series.max,
    "min": pd.Series.min,
}


class AggregateError(Exception):
    pass


def aggregate(
    hts,
    target_step,
    method,
    min_count=1,
    missing_flag="MISS",
    target_timestamp_offset=None,
):
    aggregation = Aggregation(
        source_timeseries=hts,
        target_step=target_step,
        method=method,
        min_count=min_count,
        missing_flag=missing_flag,
        target_timestamp_offset=target_timestamp_offset,
    )
    aggregation.execute()
    return aggregation.result


class Aggregation:
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)
        self.source = SourceTimeseries(self.source_timeseries)
        self.result = AggregatedTimeseries()
        self.result.time_step = self.target_step

    def execute(self):
        self.result.set_metadata(self.source_timeseries)
        try:
            self.source.normalize(self.target_step)
        except CannotInferFrequency:
            return
        self.do_aggregation()
        self.result.remove_leading_and_trailing_nans()
        self.result.add_timestamp_offset(self.target_timestamp_offset)

    def do_aggregation(self):
        self.create_resampler()
        self.get_result_values()
        self.get_result_flags()

    def create_resampler(self):
        self.resampler = self.source.data["value"].resample(
            self.result.time_step, closed="right", label="right"
        )

    def get_result_values(self):
        result_values = self.resampler.agg(methods[self.method])
        values_count = self.resampler.count()
        result_values[values_count < self.min_count] = np.nan
        self.result.data["value"] = result_values

    def get_result_flags(self):
        max_count = int(pd.Timedelta(self.result.time_step) / self.source.freq)
        values_count = self.resampler.count()
        self.result.data["flags"] = (values_count < max_count).apply(
            lambda x: self.missing_flag if x else ""
        )


class CannotInferFrequency(Exception):
    pass


attrs = ("unit", "timezone", "interval_type", "variable", "precision", "location")


class SourceTimeseries(HTimeseries):
    def __init__(self, s):
        for attr in attrs:
            setattr(self, attr, getattr(s, attr, None))
        self.data = s.data

    def normalize(self, target_step):
        """Reindex so that it has no missing records but has NaNs instead, starting from
        one before and ending in one after.
        """
        current_range = self.data.index
        try:
            self.freq = pd.tseries.frequencies.to_offset(pd.infer_freq(current_range))
            if self.freq is None:
                raise AggregateError(
                    "Can't infer time series step; maybe it's not regularized"
                )
        except ValueError:
            raise CannotInferFrequency()
        first_timestamp = (current_range[0] - pd.Timedelta("1s")).floor(target_step)
        end_timestamp = current_range[-1].ceil(target_step)
        new_range = pd.date_range(first_timestamp, end_timestamp, freq=self.freq)
        self.data = self.data.reindex(new_range)


class AggregatedTimeseries(HTimeseries):
    def set_metadata(self, source_timeseries):
        for attr in attrs:
            setattr(self, attr, getattr(source_timeseries, attr, None))
        if self.time_step not in ("1h", "1D"):
            raise AggregateError("The target step can currently only be 1h or 1D")
        if hasattr(source_timeseries, "title"):
            self.title = "Aggregated " + source_timeseries.title
        if hasattr(source_timeseries, "comment"):
            self.comment = (
                "Created by aggregating the time series that had this comment:\n\n"
                + source_timeseries.comment
            )

    def remove_leading_and_trailing_nans(self):
        while len(self.data.index) > 0 and pd.isnull(self.data["value"]).iloc[0]:
            self.data = self.data.drop(self.data.index[0])
        while len(self.data.index) > 0 and pd.isnull(self.data["value"]).iloc[-1]:
            self.data = self.data.drop(self.data.index[-1])

    def add_timestamp_offset(self, target_timestamp_offset):
        if target_timestamp_offset:
            periods = target_timestamp_offset.startswith("-") and 1 or -1
            freq = target_timestamp_offset.lstrip("-")
            self.data = self.data.shift(periods, freq=freq)


def _get_offset_in_minutes(timestamp_offset):
    m = re.match(r"(-?)(\d*)min$", timestamp_offset)
    if not m:
        raise AggregateError(
            "The target timestamp offset can currently only be a number of minutes "
            "such as 1min"
        )
    sign = m.group(1) == "-" and -1 or 1
    return sign * int(m.group(2))


class RegularizationMode(Enum):
    INSTANTANEOUS = 1
    INTERVAL = 2



from unittest import TestCase, mock

import requests

from enhydris_api_client import EnhydrisApiClient
from htimeseries import HTimeseries

from . import mock_session


class GetTokenTestCase(TestCase):
    @mock_session(
        **{
            "get.return_value.cookies": {"csrftoken": "reallysecret"},
            "post.return_value.cookies": {"acookie": "a cookie value"},
        }
    )
    def setUp(self, mock_requests_session):
        self.mock_requests_session = mock_requests_session
        self.client = EnhydrisApiClient("https://mydomain.com")
        self.client.get_token("admin", "topsecret")

    def test_makes_post_request(self):
        self.mock_requests_session.return_value.post.assert_called_once_with(
            "https://mydomain.com/api/auth/login/",
            data="username=admin&password=topsecret",
            allow_redirects=False,
        )


class GetTokenFailTestCase(TestCase):
    @mock_session(**{"post.return_value.status_code": 404})
    def test_raises_exception_on_post_failure(self, mock_requests_session):
        self.client = EnhydrisApiClient("https://mydomain.com")
        with self.assertRaises(requests.HTTPError):
            self.client.get_token("admin", "topsecret")


class GetTokenEmptyUsernameTestCase(TestCase):
    @mock_session()
    def setUp(self, mock_requests_session):
        self.mock_requests_session = mock_requests_session
        self.client = EnhydrisApiClient("https://mydomain.com")
        self.client.get_token("", "useless_password")

    def test_does_not_make_get_request(self):
        self.mock_requests_session.get.assert_not_called()

    def test_does_not_make_post_request(self):
        self.mock_requests_session.post.assert_not_called()


class UseAsContextManagerTestCase(TestCase):
    @mock_session()
    def setUp(self, mock_requests_session):
        self.mock_requests_session = mock_requests_session
        with EnhydrisApiClient("https://mydomain.com/") as api_client:
            api_client.get_station(42)

    def test_called_enter(self):
        self.mock_requests_session.return_value.__enter__.assert_called_once_with()

    def test_called_exit(self):
        self.assertEqual(
            len(self.mock_requests_session.return_value.__exit__.mock_calls), 1
        )

    def test_makes_request(self):
        self.mock_requests_session.return_value.get.assert_called_once_with(
            "https://mydomain.com/api/stations/42/"
        )


class Error400TestCase(TestCase):
    msg = "hello world"

    @mock_session(
        **{
            "get.return_value.status_code": 400,
            "get.return_value.text": "hello world",
            "post.return_value.status_code": 400,
            "post.return_value.text": "hello world",
            "put.return_value.status_code": 400,
            "put.return_value.text": "hello world",
            "patch.return_value.status_code": 400,
            "patch.return_value.text": "hello world",
            "delete.return_value.status_code": 400,
            "delete.return_value.text": "hello world",
        }
    )
    def setUp(self, m):
        self.client = EnhydrisApiClient("https://mydomain.com")

    def test_get_token(self):
        with self.assertRaisesRegex(requests.HTTPError, self.msg):
            self.client.get_token("john", "topsecret")

    def test_get_station(self):
        with self.assertRaisesRegex(requests.HTTPError, self.msg):
            self.client.get_station(42)

    def test_post_station(self):
        with self.assertRaisesRegex(requests.HTTPError, self.msg):
            self.client.post_station({})

    def test_put_station(self):
        with self.assertRaisesRegex(requests.HTTPError, self.msg):
            self.client.put_station(42, {})

    def test_patch_station(self):
        with self.assertRaisesRegex(requests.HTTPError, self.msg):
            self.client.patch_station(42, {})

    def test_delete_station(self):
        with self.assertRaisesRegex(requests.HTTPError, self.msg):
            self.client.delete_station(42)

    def test_get_timeseries(self):
        with self.assertRaisesRegex(requests.HTTPError, self.msg):
            self.client.get_timeseries(41, 42, 43)

    def test_post_timeseries(self):
        with self.assertRaisesRegex(requests.HTTPError, self.msg):
            self.client.post_timeseries(42, 43, {})

    def test_delete_timeseries(self):
        with self.assertRaisesRegex(requests.HTTPError, self.msg):
            self.client.delete_timeseries(41, 42, 43)

    def test_read_tsdata(self):
        with self.assertRaisesRegex(requests.HTTPError, self.msg):
            self.client.read_tsdata(41, 42, 43)

    def test_post_tsdata(self):
        with self.assertRaisesRegex(requests.HTTPError, self.msg):
            self.client.post_tsdata(41, 42, 43, HTimeseries())

    def test_get_ts_end_date(self):
        with self.assertRaisesRegex(requests.HTTPError, self.msg):
            self.client.get_ts_end_date(41, 42, 43)


class EnhydrisApiClientTestCase(TestCase):
    @mock.patch("requests.Session")
    def test_client_with_token(self, mock_requests_session):
        EnhydrisApiClient("https://mydomain.com/", token="test-token")
        mock_requests_session.return_value.headers.update.assert_any_call(
            {"Authorization": "token test-token"}
        )



import datetime as dt
from copy import copy
from io import StringIO
from unittest import TestCase

import requests

from enhydris_api_client import EnhydrisApiClient
from htimeseries import HTimeseries

from . import (
    AssertFrameEqualMixin,
    mock_session,
    test_timeseries_csv,
    test_timeseries_csv_bottom,
    test_timeseries_htimeseries,
)

TEST_TIMESERIES_HTS = f"Timezone=+0200\r\n\r\n{test_timeseries_csv}"


@mock_session(**{"get.return_value.text": TEST_TIMESERIES_HTS})
class ReadTsDataTestCase(TestCase, AssertFrameEqualMixin):
    url = "http://example.com/api/stations/41/timeseriesgroups/42/timeseries/43/data/"

    def _read_tsdata(self, **extra_args):
        self.client = EnhydrisApiClient("http://example.com")
        return self.client.read_tsdata(41, 42, 43, **extra_args)

    def test_makes_request(self, m):
        self._read_tsdata()
        m.return_value.get.assert_called_once_with(
            self.url,
            params={
                "fmt": "hts",
                "start_date": None,
                "end_date": None,
                "timezone": None,
            },
        )

    def test_returns_data(self, m):
        ahts = self._read_tsdata()
        self.assert_frame_equal(ahts.data, test_timeseries_htimeseries.data)

    def test_uses_timezone(self, m):
        self._read_tsdata(timezone="UTC")
        m.return_value.get.assert_called_once_with(
            self.url,
            params={
                "fmt": "hts",
                "start_date": None,
                "end_date": None,
                "timezone": "UTC",
            },
        )


class ReadTsDataWithStartAndEndDateTestCase(TestCase, AssertFrameEqualMixin):
    @mock_session(**{"get.return_value.text": TEST_TIMESERIES_HTS})
    def setUp(self, mock_requests_session):
        self.mock_requests_session = mock_requests_session
        self.client = EnhydrisApiClient("https://mydomain.com")

    def _make_request(self, start_tzinfo, end_tzinfo):
        self.data = self.client.read_tsdata(
            41,
            42,
            43,
            start_date=dt.datetime(2019, 6, 12, 0, 0, tzinfo=start_tzinfo),
            end_date=dt.datetime(2019, 6, 13, 15, 25, tzinfo=end_tzinfo),
        )

    def test_makes_request(self):
        self._make_request(dt.timezone.utc, dt.timezone.utc)
        self.mock_requests_session.return_value.get.assert_called_once_with(
            "https://mydomain.com/api/stations/41/timeseriesgroups/42/timeseries/43/"
            "data/",
            params={
                "fmt": "hts",
                "start_date": "2019-06-12T00:00:00+00:00",
                "end_date": "2019-06-13T15:25:00+00:00",
                "timezone": None,
            },
        )

    def test_returns_data(self):
        self._make_request(dt.timezone.utc, dt.timezone.utc)
        self.assert_frame_equal(self.data.data, test_timeseries_htimeseries.data)

    def test_checks_that_start_date_is_aware(self):
        with self.assertRaises(ValueError):
            self._make_request(None, dt.timezone.utc)

    def test_checks_that_end_date_is_aware(self):
        with self.assertRaises(ValueError):
            self._make_request(dt.timezone.utc, None)


class ReadEmptyTsDataTestCase(TestCase, AssertFrameEqualMixin):
    @mock_session(**{"get.return_value.text": ""})
    def test_returns_data(self, mock_requests_session):
        self.client = EnhydrisApiClient("https://mydomain.com")
        self.data = self.client.read_tsdata(41, 42, 43)
        self.assert_frame_equal(self.data.data, HTimeseries().data)


class ReadTsDataErrorTestCase(TestCase):
    @mock_session(**{"get.return_value.status_code": 404})
    def test_raises_exception_on_error(self, mock_requests_session):
        self.client = EnhydrisApiClient("https://mydomain.com")
        with self.assertRaises(requests.HTTPError):
            self.client.read_tsdata(41, 42, 43)


class PostTsDataTestCase(TestCase):
    @mock_session()
    def test_makes_request(self, mock_requests_session):
        client = EnhydrisApiClient("https://mydomain.com")
        client.post_tsdata(41, 42, 43, test_timeseries_htimeseries)
        f = StringIO()
        data = copy(test_timeseries_htimeseries.data)
        data.index = data.index.tz_convert("UTC")
        data.to_csv(f, header=False)
        mock_requests_session.return_value.post.assert_called_once_with(
            "https://mydomain.com/api/stations/41/timeseriesgroups/42/timeseries/43/"
            "data/",
            data={"timeseries_records": f.getvalue(), "timezone": "UTC"},
        )

    @mock_session(**{"post.return_value.status_code": 404})
    def test_raises_exception_on_error(self, mock_requests_session):
        client = EnhydrisApiClient("https://mydomain.com")
        with self.assertRaises(requests.HTTPError):
            client.post_tsdata(41, 42, 43, test_timeseries_htimeseries)


@mock_session(**{"get.return_value.text": test_timeseries_csv_bottom})
class GetTsEndDateTestCase(TestCase):
    url = "http://mydom.com/api/stations/41/timeseriesgroups/42/timeseries/43/bottom/"

    def _get_ts_end_date(self, **extra_args):
        self.client = EnhydrisApiClient("http://mydom.com")
        return self.client.get_ts_end_date(41, 42, 43, **extra_args)

    def test_makes_request(self, m):
        self._get_ts_end_date()
        m.return_value.get.assert_called_once_with(self.url, params={"timezone": None})

    def test_returns_date(self, m):
        result = self._get_ts_end_date()
        self.assertEqual(result, dt.datetime(2014, 1, 5, 8, 0))

    def test_uses_timezone(self, m):
        self._get_ts_end_date(timezone="Etc/GMT-2")
        m.return_value.get.assert_called_once_with(
            self.url, params={"timezone": "Etc/GMT-2"}
        )


class GetTsEndDateErrorTestCase(TestCase):
    @mock_session(**{"get.return_value.status_code": 404})
    def test_checks_response_code(self, mock_requests_session):
        client = EnhydrisApiClient("https://mydomain.com")
        with self.assertRaises(requests.HTTPError):
            client.get_ts_end_date(41, 42, 43)


class GetTsEndDateEmptyTestCase(TestCase):
    @mock_session(**{"get.return_value.text": ""})
    def test_returns_date(self, mock_requests_session):
        client = EnhydrisApiClient("https://mydomain.com")
        date = client.get_ts_end_date(41, 42, 43)
        self.assertIsNone(date)



from unittest import TestCase

import requests

from enhydris_api_client import EnhydrisApiClient

from . import mock_session


class GetStationTestCase(TestCase):
    @mock_session(**{"get.return_value.json.return_value": {"hello": "world"}})
    def setUp(self, mock_requests_session):
        self.mock_requests_session = mock_requests_session
        self.client = EnhydrisApiClient("https://mydomain.com")
        self.data = self.client.get_station(42)

    def test_makes_request(self):
        self.mock_requests_session.return_value.get.assert_called_once_with(
            "https://mydomain.com/api/stations/42/"
        )

    def test_returns_data(self):
        self.assertEqual(self.data, {"hello": "world"})


class PostStationTestCase(TestCase):
    @mock_session(**{"post.return_value.json.return_value": {"id": 42}})
    def setUp(self, mock_requests_session):
        self.mock_requests_session = mock_requests_session
        self.client = EnhydrisApiClient("https://mydomain.com")
        self.data = self.client.post_station(data={"location": "Syria"})

    def test_makes_request(self):
        self.mock_requests_session.return_value.post.assert_called_once_with(
            "https://mydomain.com/api/stations/", data={"location": "Syria"}
        )

    def test_returns_id(self):
        self.assertEqual(self.data, 42)


class PutStationTestCase(TestCase):
    @mock_session()
    def setUp(self, mock_requests_session):
        self.mock_requests_session = mock_requests_session
        self.client = EnhydrisApiClient("https://mydomain.com")
        self.client.put_station(42, data={"location": "Syria"})

    def test_makes_request(self):
        self.mock_requests_session.return_value.put.assert_called_once_with(
            "https://mydomain.com/api/stations/42/", data={"location": "Syria"}
        )


class PatchStationTestCase(TestCase):
    @mock_session()
    def setUp(self, mock_requests_session):
        self.mock_requests_session = mock_requests_session
        self.client = EnhydrisApiClient("https://mydomain.com")
        self.client.patch_station(42, data={"location": "Syria"})

    def test_makes_request(self):
        self.mock_requests_session.return_value.patch.assert_called_once_with(
            "https://mydomain.com/api/stations/42/", data={"location": "Syria"}
        )


class DeleteStationTestCase(TestCase):
    @mock_session()
    def test_makes_request(self, mock_requests_session):
        self.client = EnhydrisApiClient("https://mydomain.com")
        self.client.delete_station(42)
        mock_requests_session.return_value.delete.assert_called_once_with(
            "https://mydomain.com/api/stations/42/"
        )

    @mock_session(**{"delete.return_value.status_code": 404})
    def test_raises_exception_on_error(self, mock_requests_delete):
        self.client = EnhydrisApiClient("https://mydomain.com")
        with self.assertRaises(requests.HTTPError):
            self.client.delete_station(42)



import datetime as dt
import math
import textwrap
from io import StringIO
from unittest import TestCase

try:
    from zoneinfo import ZoneInfo
except ImportError:
    from backports.zoneinfo import ZoneInfo

import numpy as np

from haggregate import RegularizationMode, RegularizeError, regularize
from htimeseries import HTimeseries


class BadTimeStepTestCase(TestCase):
    def test_unspecified_time_step(self):
        ts = HTimeseries()
        msg = "The source time series does not specify a time step"
        with self.assertRaisesRegex(RegularizeError, msg):
            regularize(ts, mode=RegularizationMode.INTERVAL)

    def test_malformed_time_step(self):
        ts = HTimeseries()
        ts.time_step = "hello"
        msg = (
            "The time step is malformed or is specified in months. Only time steps "
            "specified in minutes, hours or days are supported."
        )
        with self.assertRaisesRegex(RegularizeError, msg):
            regularize(ts, mode=RegularizationMode.INTERVAL)

    def test_malformed_time_step2(self):
        ts = HTimeseries()
        ts.time_step = "5M"
        msg = (
            "The time step is malformed or is specified in months. Only time steps "
            "specified in minutes, hours or days are supported."
        )
        with self.assertRaisesRegex(RegularizeError, msg):
            regularize(ts, mode=RegularizationMode.INTERVAL)


class RegularizeTestCase(TestCase):
    def setUp(self):
        input = textwrap.dedent(
            """\
            2008-02-07 10:30,10.71,FLAG1
            2008-02-07 10:41,10.93,FLAG2
            2008-02-07 10:50,11.10,
            """
        )
        ts = HTimeseries(StringIO(input), default_tzinfo=ZoneInfo("Etc/GMT-2"))
        ts.time_step = "10min"
        self.result = regularize(ts, mode=RegularizationMode.INTERVAL)

    def test_length(self):
        self.assertEqual(len(self.result.data), 3)

    def test_timestamps_are_aware(self):
        self.assertEqual(self.result.data.index[0].utcoffset(), dt.timedelta(hours=2))

    def test_value_1(self):
        self.assertAlmostEqual(
            self.result.data.loc["2008-02-07 10:30:00+0200"].value, 10.71
        )

    def test_value_2(self):
        self.assertAlmostEqual(
            self.result.data.loc["2008-02-07 10:40:00+0200"].value, 10.93
        )

    def test_value_3(self):
        self.assertAlmostEqual(
            self.result.data.loc["2008-02-07 10:50:00+0200"].value, 11.10
        )

    def test_flags_1(self):
        self.assertEqual(
            self.result.data.loc["2008-02-07 10:30:00+0200"]["flags"], "FLAG1"
        )

    def test_flags_2(self):
        self.assertEqual(
            self.result.data.loc["2008-02-07 10:40:00+0200"]["flags"],
            "FLAG2 DATEINSERT",
        )

    def test_flags_3(self):
        self.assertEqual(self.result.data.loc["2008-02-07 10:50:00+0200"]["flags"], "")


class RegularizeFirstRecordTestCase(TestCase):
    def setUp(self):
        input = textwrap.dedent(
            """\
            2008-02-07 10:31,10.71,
            2008-02-07 10:40,10.93,
            2008-02-07 10:50,11.10,
            """
        )
        ts = HTimeseries(StringIO(input), default_tzinfo=ZoneInfo("Etc/GMT-2"))
        ts.time_step = "10min"
        self.result = regularize(ts, mode=RegularizationMode.INTERVAL)

    def test_length(self):
        self.assertEqual(len(self.result.data), 3)

    def test_value_1(self):
        self.assertAlmostEqual(
            self.result.data.loc["2008-02-07 10:30:00+0200"].value, 10.71
        )

    def test_value_2(self):
        self.assertAlmostEqual(
            self.result.data.loc["2008-02-07 10:40:00+0200"].value, 10.93
        )

    def test_value_3(self):
        self.assertAlmostEqual(
            self.result.data.loc["2008-02-07 10:50:00+0200"].value, 11.10
        )


class RegularizeLastRecordTestCase(TestCase):
    def setUp(self):
        input = textwrap.dedent(
            """\
            2008-02-07 10:30,10.71,
            2008-02-07 10:40,10.93,
            2008-02-07 10:49,11.10,
            """
        )
        ts = HTimeseries(StringIO(input), default_tzinfo=ZoneInfo("Etc/GMT-2"))
        ts.time_step = "10min"
        self.result = regularize(ts, mode=RegularizationMode.INTERVAL)

    def test_length(self):
        self.assertEqual(len(self.result.data), 3)

    def test_value_1(self):
        self.assertAlmostEqual(
            self.result.data.loc["2008-02-07 10:30:00+0200"].value, 10.71
        )

    def test_value_2(self):
        self.assertAlmostEqual(
            self.result.data.loc["2008-02-07 10:40:00+0200"].value, 10.93
        )

    def test_value_3(self):
        self.assertAlmostEqual(
            self.result.data.loc["2008-02-07 10:50:00+0200"].value, 11.10
        )


class RegularizeNullRecordTestCase(TestCase):
    def setUp(self):
        input = textwrap.dedent(
            """\
            2008-02-07 10:30,10.71,
            2008-02-07 10:51,11.10,
            2008-02-07 11:00,10.93,
            """
        )
        ts = HTimeseries(StringIO(input), default_tzinfo=ZoneInfo("Etc/GMT-2"))
        ts.time_step = "10min"
        self.result = regularize(ts, mode=RegularizationMode.INTERVAL)

    def test_length(self):
        self.assertEqual(len(self.result.data), 4)

    def test_value_1(self):
        self.assertAlmostEqual(
            self.result.data.loc["2008-02-07 10:30:00+0200"].value, 10.71
        )

    def test_value_2(self):
        self.assertTrue(
            np.isnan(self.result.data.loc["2008-02-07 10:40:00+0200"].value)
        )

    def test_value_3(self):
        self.assertAlmostEqual(
            self.result.data.loc["2008-02-07 10:50:00+0200"].value, 11.10
        )

    def test_value_4(self):
        self.assertAlmostEqual(
            self.result.data.loc["2008-02-07 11:00:00+0200"].value, 10.93
        )


class RegularizeEmptyTestCase(TestCase):
    def setUp(self):
        ts = HTimeseries()
        ts.time_step = "10min"
        self.result = regularize(ts, mode=RegularizationMode.INTERVAL)

    def test_length(self):
        self.assertEqual(len(self.result.data), 0)


class RegularizeWithNullRecordTestCase(TestCase):
    def setUp(self):
        input = textwrap.dedent(
            """\
            2008-02-07 10:30,10.71,FLAG1
            2008-02-07 10:40,,
            2008-02-07 10:41,10.93,FLAG2
            2008-02-07 10:50,11.10,
            """
        )
        self.ts = HTimeseries(StringIO(input), default_tzinfo=ZoneInfo("Etc/GMT-2"))
        self.ts.time_step = "10min"

    def test_interval(self):
        result = regularize(self.ts, mode=RegularizationMode.INTERVAL)
        self.assertTrue(math.isnan(result.data.loc["2008-02-07 10:40"].value))

    def test_instantaneous(self):
        result = regularize(self.ts, mode=RegularizationMode.INSTANTANEOUS)
        self.assertAlmostEqual(result.data.loc["2008-02-07 10:40"].value, 10.93)


class NearestTestCase(TestCase):
    def setUp(self):
        input = textwrap.dedent(
            """\
            2008-02-07 10:30,10.71,FLAG1
            2008-02-07 10:38,11.93,FLAG2
            2008-02-07 10:41,10.93,FLAG2
            2008-02-07 10:50,11.10,
            """
        )
        self.ts = HTimeseries(StringIO(input), default_tzinfo=ZoneInfo("Etc/GMT-2"))
        self.ts.time_step = "10min"

    def test_interval(self):
        result = regularize(self.ts, mode=RegularizationMode.INTERVAL)
        self.assertTrue(math.isnan(result.data.loc["2008-02-07 10:40"].value))

    def test_instantaneous(self):
        result = regularize(self.ts, mode=RegularizationMode.INSTANTANEOUS)
        self.assertAlmostEqual(result.data.loc["2008-02-07 10:40"].value, 10.93)


class SetsMetadataTestCase(TestCase):
    def setUp(self):
        input = textwrap.dedent(
            """\
            2008-02-07 10:30,10.71,FLAG1
            2008-02-07 10:41,10.93,FLAG2
            2008-02-07 10:50,11.10,
            """
        )
        self.ts = HTimeseries(StringIO(input), default_tzinfo=ZoneInfo("Etc/GMT-2"))
        self.ts.time_step = "10min"
        self.ts.title = "hello"
        self.ts.precision = 1
        self.ts.comment = "world"
        self.ts.timezone = "EET (+0200)"
        self.result = regularize(self.ts, mode=RegularizationMode.INTERVAL)

    def test_sets_title(self):
        self.assertEqual(self.result.title, "Regularized hello")

    def test_sets_precision(self):
        self.assertEqual(self.result.precision, 1)

    def test_sets_comment(self):
        self.assertEqual(
            self.result.comment,
            "Created by regularizing step of timeseries "
            "that had this comment:\n\nworld",
        )

    def test_sets_time_step(self):
        self.assertEqual(self.result.time_step, "10min")

    def test_sets_timezone(self):
        self.assertEqual(self.result.timezone, "EET (+0200)")



import configparser
import datetime as dt
import logging
import os
import traceback
from glob import glob
from io import StringIO

import click
import iso8601
from osgeo import gdal, ogr, osr
from simpletail import ropen

from hspatial import create_ogr_layer_from_timeseries, h_integrate, idw
from htimeseries import HTimeseries, TzinfoFromString
from pthelma._version import __version__

gdal.UseExceptions()


class WrongValueError(configparser.Error):
    pass


class App:
    def __init__(self, configfilename):
        self.configfilename = configfilename

    def run(self):
        self.config = AppConfig(self.configfilename)
        self.config.read()
        self._setup_logger()
        self._execute_with_error_handling()

    def _execute_with_error_handling(self):
        self.logger.info("Starting spatialize, " + dt.datetime.today().isoformat())
        try:
            self._execute()
        except Exception as e:
            self.logger.error(str(e))
            self.logger.debug(traceback.format_exc())
            self.logger.info(
                "spatialize terminated with error, " + dt.datetime.today().isoformat()
            )
            raise click.ClickException(str(e))
        else:
            self.logger.info("Finished spatialize, " + dt.datetime.today().isoformat())

    def _setup_logger(self):
        self.logger = logging.getLogger("spatialize")
        self._set_logger_handler()
        self.logger.setLevel(self.config.loglevel.upper())

    def _set_logger_handler(self):
        if getattr(self.config, "logfile", None):
            self.logger.addHandler(logging.FileHandler(self.config.logfile))
        else:
            self.logger.addHandler(logging.StreamHandler())

    def _get_last_dates(self, filename, n):
        """
        Assuming specified file contains a time series, scan it from the bottom
        and return the list of the n last dates (may be less than n if the time
        series is too small). 'filename' is used in error messages.
        """
        # Get the time zone
        with open(filename) as fp:
            for line in fp:
                if line.startswith("Timezone") or (line and line[0] in "0123456789"):
                    break
        if not line.startswith("Timezone"):
            raise click.ClickException("{} does not contain Timezone".format(filename))
        zonestr = line.partition("=")[2].strip()
        timezone = TzinfoFromString(zonestr)

        result = []
        previous_line_was_empty = False
        with ropen(filename) as fp:
            for i, line in enumerate(fp):
                if i >= n:
                    break
                line = line.strip()

                # Ignore empty lines
                if not line:
                    previous_line_was_empty = True
                    continue

                # Is the line in the form of an ini file configuration line?
                items = line.split("=")
                if len(items) and ("," not in items[0]) and previous_line_was_empty:
                    break  # Yes; we reached the start of the file

                previous_line_was_empty = False

                datestring = line.split(",")[0]
                try:
                    result.insert(
                        0, iso8601.parse_date(datestring, default_timezone=timezone)
                    )
                except iso8601.ParseError as e:
                    raise click.ClickException(
                        str(e)
                        + " (file {}, {} lines from the end)".format(filename, i + 1)
                    )
        return result

    @property
    def _dates_to_calculate(self):
        """
        Generator that yields the dates for which h_integrate should be run;
        this is the latest list of dates such that:
        * At least one of the time series has data
        * The length of the list is the 'number_of_output_files' configuration
          option (maybe less if the time series don't have enough data yet).
        """
        n = self.config.number_of_output_files
        dates = set()
        for filename in self.config.files:
            dates |= set(self._get_last_dates(filename, n))
        dates = list(dates)
        dates.sort()
        dates = dates[-n:]
        for d in dates:
            yield d

    @property
    def _time_step(self):
        """
        Return time step of all time series. If time step is not the same
        for all time series, raises exception.
        """
        time_step = None
        for filename in self.config.files:
            with open(filename, newline="\n") as f:
                t = HTimeseries(f, start_date="0001-01-01 00:00")
            item_time_step = t.time_step
            if time_step and (item_time_step != time_step):
                raise click.ClickException("Not all time series have the same step")
            time_step = item_time_step
        return time_step

    @property
    def _date_fmt(self):
        """
        Determine date_fmt based on time series time step.
        """
        if self._time_step.endswith("min") or self._time_step.endswith("h"):
            return "%Y-%m-%d %H:%M%z"
        elif self._time_step.endswith("D"):
            return "%Y-%m-%d"
        elif self._time_step.endswith("M"):
            return "%Y-%m"
        elif self._time_step.endswith("Y"):
            return "%Y"
        raise click.ClickException("Can't use time step " + str(self._time_step))

    def _delete_obsolete_files(self):
        """
        Delete all tif files produced in the past except the last N,
        where N is the 'number_of_output_files' configuration option.
        """
        pattern = os.path.join(
            self.config.output_dir, "{}-*.tif".format(self.config.filename_prefix)
        )
        files = glob(pattern)
        files.sort()
        for filename in files[: -self.config.number_of_output_files]:
            os.remove(filename)

    def _execute(self):
        # Create stations layer
        stations = ogr.GetDriverByName("memory").CreateDataSource("stations")
        stations_layer = create_ogr_layer_from_timeseries(
            self.config.files, self.config.epsg, stations
        )

        # Get mask
        mask = gdal.Open(self.config.mask)

        # Setup integration method
        if self.config.method == "idw":
            funct = idw
            kwargs = {"alpha": self.config.alpha}
        else:
            assert False

        for date in self._dates_to_calculate:
            self.logger.info("Processing date " + date.isoformat())
            h_integrate(
                mask,
                stations_layer,
                date,
                os.path.join(self.config.output_dir, self.config.filename_prefix),
                self._date_fmt,
                funct,
                kwargs,
            )
        self._delete_obsolete_files()


class AppConfig:
    config_file_options = {
        "logfile": {"fallback": ""},
        "loglevel": {"fallback": "warning"},
        "mask": {},
        "epsg": {},
        "output_dir": {},
        "filename_prefix": {},
        "number_of_output_files": {},
        "method": {},
        "alpha": {"fallback": "1"},
        "files": {},
    }

    def __init__(self, configfilename):
        self.configfilename = configfilename

    def read(self):
        try:
            self._parse_config()
        except (OSError, configparser.Error) as e:
            raise click.ClickException(str(e))

    def _parse_config(self):
        self._read_config_file()
        self._get_config_options()
        self._parse_config_options()

    def _read_config_file(self):
        self.config = configparser.ConfigParser(interpolation=None)
        try:
            self._read_config_file_assuming_it_has_section_headers()
        except configparser.MissingSectionHeaderError:
            self._read_config_file_without_sections()

    def _read_config_file_assuming_it_has_section_headers(self):
        with open(self.configfilename) as f:
            self.config.read_file(f)

    def _read_config_file_without_sections(self):
        with open(self.configfilename) as f:
            configuration = "[General]\n" + f.read()
        self.config.read_file(StringIO(configuration))

    def _get_config_options(self):
        self.options = {
            opt: self.config.get("General", opt, **kwargs)
            for opt, kwargs in self.config_file_options.items()
        }
        for key, value in self.options.items():
            setattr(self, key, value)

    def _parse_config_options(self):
        self._parse_log_level()
        self._parse_files()
        self._check_method()
        self._parse_epsg()
        self._parse_number_of_output_files()

    def _parse_log_level(self):
        log_levels = ("ERROR", "WARNING", "INFO", "DEBUG")
        self.loglevel = self.loglevel.upper()
        if self.loglevel not in log_levels:
            raise WrongValueError("loglevel must be one of " + ", ".join(log_levels))

    def _parse_files(self):
        self.files = self.files.split("\n")

    def _check_method(self):
        # Check method
        if self.method != "idw":
            raise WrongValueError('Option "method" can currently only be idw')
        # Check alpha
        try:
            self.alpha = float(self.alpha)
        except ValueError:
            raise WrongValueError('Option "alpha" must be a number')

    def _parse_epsg(self):
        try:
            self.epsg = int(self.epsg)
        except ValueError:
            raise WrongValueError('Option "epsg" must be an integer')
        srs = osr.SpatialReference()
        result = srs.ImportFromEPSG(self.epsg)
        if result:
            raise WrongValueError(
                "An error occurred when trying to use epsg={}".format(self.epsg)
            )

    def _parse_number_of_output_files(self):
        try:
            self.number_of_output_files = int(self.number_of_output_files)
        except ValueError:
            raise WrongValueError('Option "number_of_output_files" must be an integer')


@click.command()
@click.argument("configfile")
@click.version_option(
    version=__version__, message="%(prog)s from pthelma v.%(version)s"
)
def main(configfile):
    """Spatial integration"""

    app = App(configfile)
    app.run()



import datetime as dt
import os
import shutil
import tempfile
import textwrap
from pathlib import Path
from unittest import TestCase
from unittest.mock import patch

import click
from click.testing import CliRunner
from osgeo import gdal, osr

from hspatial import cli
from htimeseries import TzinfoFromString

gdal.UseExceptions()


def create_geotiff_file(filename, value):
    geo_transform = (-16.25, 1.0, 0, 16.217, 0, 1.0)
    wgs84 = osr.SpatialReference()
    wgs84.ImportFromEPSG(4326)
    f = gdal.GetDriverByName("GTiff").Create(filename, 2, 1, 1, gdal.GDT_Float32)
    try:
        f.SetGeoTransform(geo_transform)
        f.SetProjection(wgs84.ExportToWkt())
        f.GetRasterBand(1).WriteArray(value)
    finally:
        # Close the dataset
        f = None


class NonExistentConfigFileTestCase(TestCase):
    def setUp(self):
        runner = CliRunner(mix_stderr=False)
        self.result = runner.invoke(cli.main, ["nonexistent.conf"])

    def test_exit_status(self):
        self.assertEqual(self.result.exit_code, 1)

    def test_error_message(self):
        self.assertIn(
            "No such file or directory: 'nonexistent.conf'", self.result.stderr
        )


class ConfigurationTestCase(TestCase):
    def setUp(self):
        self.tempdir = tempfile.mkdtemp()
        self.configfilename = os.path.join(self.tempdir, "hspatial.conf")

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    def test_missing_mask_parameter_raises_error(self):
        with open(self.configfilename, "w") as f:
            f.write(
                textwrap.dedent(
                    f"""\
                    logfile = {os.path.join(self.tempdir, "logfile")}
                    epsg = 2100
                    output_dir = /var/opt/hspatial
                    filename_prefix = rainfall
                    number_of_output_files = 24
                    method = idw
                    files = /var/opt/timeseries/inputfile1.hts
                            /var/opt/timeseries/inputfile2.hts
                            /var/opt/timeseries/inputfile3.hts
                    """
                )
            )
        msg = "No option 'mask'"
        with self.assertRaisesRegex(click.ClickException, msg):
            cli.App(self.configfilename).run()

    def test_nonexistent_log_level_raises_error(self):
        with open(self.configfilename, "w") as f:
            f.write(
                textwrap.dedent(
                    f"""\
                    logfile = {os.path.join(self.tempdir, "logfile")}
                    loglevel = HELLO
                    mask = /etc/hspatial/mask.tif
                    epsg = 2100
                    output_dir = /var/opt/hspatial
                    filename_prefix = rainfall
                    number_of_output_files = 24
                    method = idw
                    files = /var/opt/timeseries/inputfile1.hts
                            /var/opt/timeseries/inputfile2.hts
                            /var/opt/timeseries/inputfile3.hts
                    """
                )
            )
        msg = "loglevel must be one of ERROR, WARNING, INFO, DEBUG"
        with self.assertRaisesRegex(click.ClickException, msg):
            cli.App(self.configfilename).run()

    def test_missing_epsg_parameter_raises_error(self):
        with open(self.configfilename, "w") as f:
            f.write(
                textwrap.dedent(
                    f"""\
                    logfile = {os.path.join(self.tempdir, "logfile")}
                    mask = /etc/hspatial/mask.tif
                    output_dir = /var/opt/hspatial
                    filename_prefix = rainfall
                    number_of_output_files = 24
                    method = idw
                    files = /var/opt/timeseries/inputfile1.hts
                            /var/opt/timeseries/inputfile2.hts
                            /var/opt/timeseries/inputfile3.hts
                    """
                )
            )
        msg = "No option 'epsg'"
        with self.assertRaisesRegex(click.ClickException, msg):
            cli.App(self.configfilename).run()

    def test_wrong_epsg_parameter_raises_error(self):
        with open(self.configfilename, "w") as f:
            f.write(
                textwrap.dedent(
                    f"""\
                    logfile = {os.path.join(self.tempdir, "logfile")}
                    mask = /etc/hspatial/mask.tif
                    epsg = hello
                    output_dir = /var/opt/hspatial
                    filename_prefix = rainfall
                    number_of_output_files = 24
                    method = idw
                    files = /var/opt/timeseries/inputfile1.hts
                            /var/opt/timeseries/inputfile2.hts
                            /var/opt/timeseries/inputfile3.hts
                    """
                )
            )
        msg = 'Option "epsg" must be an integer'
        with self.assertRaisesRegex(click.ClickException, msg):
            cli.App(self.configfilename).run()

    def test_missing_filename_prefix_parameter_raises_error(self):
        with open(self.configfilename, "w") as f:
            f.write(
                textwrap.dedent(
                    f"""\
                    logfile = {os.path.join(self.tempdir, "logfile")}
                    mask = /etc/hspatial/mask.tif
                    epsg = 2100
                    output_dir = /var/opt/hspatial
                    number_of_output_files = 24
                    method = idw
                    files = /var/opt/timeseries/inputfile1.hts
                            /var/opt/timeseries/inputfile2.hts
                            /var/opt/timeseries/inputfile3.hts
                    """
                )
            )
        msg = "No option 'filename_prefix'"
        with self.assertRaisesRegex(click.ClickException, msg):
            cli.App(self.configfilename).run()

    def test_missing_output_dir_parameter_raises_error(self):
        with open(self.configfilename, "w") as f:
            f.write(
                textwrap.dedent(
                    f"""\
                    logfile = {os.path.join(self.tempdir, "logfile")}
                    mask = /etc/hspatial/mask.tif
                    epsg = 2100
                    filename_prefix = rainfall
                    number_of_output_files = 24
                    method = idw
                    files = /var/opt/timeseries/inputfile1.hts
                            /var/opt/timeseries/inputfile2.hts
                            /var/opt/timeseries/inputfile3.hts
                    """
                )
            )
        msg = "No option 'output_dir'"
        with self.assertRaisesRegex(click.ClickException, msg):
            cli.App(self.configfilename).run()

    def test_missing_number_of_output_files_parameter_raises_error(self):
        with open(self.configfilename, "w") as f:
            f.write(
                textwrap.dedent(
                    f"""\
                    logfile = {os.path.join(self.tempdir, "logfile")}
                    mask = /etc/hspatial/mask.tif
                    epsg = 2100
                    output_dir = /var/opt/hspatial
                    filename_prefix = rainfall
                    method = idw
                    files = /var/opt/timeseries/inputfile1.hts
                            /var/opt/timeseries/inputfile2.hts
                            /var/opt/timeseries/inputfile3.hts
                    """
                )
            )
        msg = "No option 'number_of_output_files'"
        with self.assertRaisesRegex(click.ClickException, msg):
            cli.App(self.configfilename).run()

    def test_wrong_number_of_output_files_parameter_raises_error(self):
        with open(self.configfilename, "w") as f:
            f.write(
                textwrap.dedent(
                    f"""\
                    logfile = {os.path.join(self.tempdir, "logfile")}
                    mask = /etc/hspatial/mask.tif
                    epsg = 2100
                    output_dir = /var/opt/hspatial
                    filename_prefix = rainfall
                    method = idw
                    number_of_output_files = hello
                    files = /var/opt/timeseries/inputfile1.hts
                            /var/opt/timeseries/inputfile2.hts
                            /var/opt/timeseries/inputfile3.hts
                    """
                )
            )
        msg = 'Option "number_of_output_files" must be an integer'
        with self.assertRaisesRegex(click.ClickException, msg):
            cli.App(self.configfilename).run()

    def test_missing_method_raises_error(self):
        with open(self.configfilename, "w") as f:
            f.write(
                textwrap.dedent(
                    f"""\
                    logfile = {os.path.join(self.tempdir, "logfile")}
                    mask = /etc/hspatial/mask.tif
                    epsg = 2100
                    output_dir = /var/opt/hspatial
                    filename_prefix = rainfall
                    number_of_output_files = 24
                    files = /var/opt/timeseries/inputfile1.hts
                            /var/opt/timeseries/inputfile2.hts
                            /var/opt/timeseries/inputfile3.hts
                    """
                )
            )
        msg = "No option 'method'"
        with self.assertRaisesRegex(click.ClickException, msg):
            cli.App(self.configfilename).run()

    def test_missing_files_parameter_raises_error(self):
        with open(self.configfilename, "w") as f:
            f.write(
                textwrap.dedent(
                    f"""\
                    logfile = {os.path.join(self.tempdir, "logfile")}
                    mask = /etc/hspatial/mask.tif
                    epsg = 2100
                    output_dir = /var/opt/hspatial
                    filename_prefix = rainfall
                    method = idw
                    number_of_output_files = 24
                    """
                )
            )
        msg = "No option 'files'"
        with self.assertRaisesRegex(click.ClickException, msg):
            cli.App(self.configfilename).run()

    def test_wrong_alpha_parameter_raises_error(self):
        with open(self.configfilename, "w") as f:
            f.write(
                textwrap.dedent(
                    f"""\
                    logfile = {os.path.join(self.tempdir, "logfile")}
                    mask = /etc/hspatial/mask.tif
                    epsg = 2100
                    output_dir = /var/opt/hspatial
                    filename_prefix = rainfall
                    method = idw
                    number_of_output_files = 24
                    alpha = hello
                    files = /var/opt/timeseries/inputfile1.hts
                            /var/opt/timeseries/inputfile2.hts
                            /var/opt/timeseries/inputfile3.hts
                    """
                )
            )
        msg = 'Option "alpha" must be a number'
        with self.assertRaisesRegex(click.ClickException, msg):
            cli.App(self.configfilename).run()

    @patch("hspatial.cli.App._execute")
    def test_correct_configuration_executes(self, m):
        with open(self.configfilename, "w") as f:
            f.write(
                textwrap.dedent(
                    f"""\
                    logfile = {os.path.join(self.tempdir, "logfile")}
                    mask = /etc/hspatial/mask.tif
                    epsg = 2100
                    output_dir = /var/opt/hspatial
                    filename_prefix = rainfall
                    method = idw
                    number_of_output_files = 24
                    files = /var/opt/timeseries/inputfile1.hts
                            /var/opt/timeseries/inputfile2.hts
                            /var/opt/timeseries/inputfile3.hts
                    """
                )
            )
        cli.App(self.configfilename).run()
        m.assert_called_once_with()

    @patch("hspatial.cli.App._execute")
    def test_creates_log_file(self, *args):
        logfilename = os.path.join(self.tempdir, "hspatial.log")
        with open(self.configfilename, "w") as f:
            f.write(
                textwrap.dedent(
                    f"""\
                    logfile = {logfilename}
                    mask = /etc/hspatial/mask.tif
                    epsg = 2100
                    output_dir = /var/opt/hspatial
                    filename_prefix = rainfall
                    method = idw
                    number_of_output_files = 24
                    files = /var/opt/timeseries/inputfile1.hts
                            /var/opt/timeseries/inputfile2.hts
                            /var/opt/timeseries/inputfile3.hts
                    """
                )
            )
        cli.App(self.configfilename).run()
        self.assertTrue(os.path.exists(logfilename))


def _create_test_data(filename1, filename2):
    with open(filename1, "w") as f:
        f.write(
            textwrap.dedent(
                """\
                Location=23.78743 37.97385 4326
                Timezone=+0200
                Time_step=60,0

                2014-04-30 11:00,18.3,
                2014-04-30 13:00,20.4,
                2014-04-30 14:00,21.4,
                2014-04-30 15:00,22.4,
                """
            )
        )
    with open(filename2, "w") as f:
        f.write(
            textwrap.dedent(
                """\
                Location=24.56789 38.76543 4326
                Timezone=EET (+0200)
                Time_step=60,0

                2014-04-30 11:00,18.3,
                2014-04-30 12:00,19.3,
                2014-04-30 13:00,20.4,
                2014-04-30 14:00,21.4,
                """
            )
        )


class AppTestCase(TestCase):
    def setUp(self):
        self.tempdir = tempfile.mkdtemp()
        self.output_dir = os.path.join(self.tempdir, "output")
        self.config_file = os.path.join(self.tempdir, "spatialize.conf")
        os.mkdir(self.output_dir)
        self.mask_file = os.path.join(self.tempdir, "mask.tif")
        self.filenames = [os.path.join(self.tempdir, x) for x in ("ts1", "ts2")]
        _create_test_data(self.filenames[0], self.filenames[1])
        self._prepare_config_file()

    def _prepare_config_file(self, number_of_output_files=24):
        with open(self.config_file, "w") as f:
            f.write(
                textwrap.dedent(
                    f"""\
                    logfile = {os.path.join(self.tempdir, "logfile")}
                    mask = {self.mask_file}
                    epsg = 2100
                    output_dir = {self.output_dir}
                    filename_prefix = rainfall
                    number_of_output_files = {number_of_output_files}
                    method = idw
                    files = {self.filenames[0]}
                            {self.filenames[1]}
                    """
                )
            )

    def _create_mask_file(self):
        mask_filename = os.path.join(self.tempdir, "mask.tif")
        mask = gdal.GetDriverByName("GTiff").Create(
            mask_filename, 640, 480, 1, gdal.GDT_Float32
        )
        mask.SetGeoTransform((23, 0.01, 0, 39, 0, -0.01))
        mask.SetProjection(
            'GEOGCS["WGS 84",DATUM["WGS_1984",SPHEROID["WGS 84",6378137,'
            '298.257223563,AUTHORITY["EPSG","7030"]],AUTHORITY["EPSG","6326"'
            ']],PRIMEM["Greenwich",0,AUTHORITY["EPSG","8901"]],UNIT["degree",'
            '0.01745329251994328,AUTHORITY["EPSG","9122"]],AUTHORITY["EPSG",'
            '"4326"]]'
        )
        mask = None

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    def test_get_last_dates(self):
        application = cli.App(self.config_file)
        tzinfo = TzinfoFromString("+0200")
        self.assertEqual(
            application._get_last_dates(self.filenames[0], 2),
            [
                dt.datetime(2014, 4, 30, 14, 0, tzinfo=tzinfo),
                dt.datetime(2014, 4, 30, 15, 0, tzinfo=tzinfo),
            ],
        )
        self.assertEqual(
            application._get_last_dates(self.filenames[0], 20),
            [
                dt.datetime(2014, 4, 30, 11, 0, tzinfo=tzinfo),
                dt.datetime(2014, 4, 30, 13, 0, tzinfo=tzinfo),
                dt.datetime(2014, 4, 30, 14, 0, tzinfo=tzinfo),
                dt.datetime(2014, 4, 30, 15, 0, tzinfo=tzinfo),
            ],
        )

    @patch("hspatial.cli.App._execute")
    def test_dates_to_calculate(self, *args):
        application = cli.App(self.config_file)
        application.run()
        tzinfo = TzinfoFromString("+0200")

        # Check for number_of_output_files=24
        dates = []
        for d in application._dates_to_calculate:
            dates.append(d)
        self.assertEqual(
            dates,
            [
                dt.datetime(2014, 4, 30, 11, 0, tzinfo=tzinfo),
                dt.datetime(2014, 4, 30, 12, 0, tzinfo=tzinfo),
                dt.datetime(2014, 4, 30, 13, 0, tzinfo=tzinfo),
                dt.datetime(2014, 4, 30, 14, 0, tzinfo=tzinfo),
                dt.datetime(2014, 4, 30, 15, 0, tzinfo=tzinfo),
            ],
        )

        # Check for number_of_output_files=2
        application.config.number_of_output_files = 2
        dates = []
        for d in application._dates_to_calculate:
            dates.append(d)
        self.assertEqual(
            dates,
            [
                dt.datetime(2014, 4, 30, 14, 0, tzinfo=tzinfo),
                dt.datetime(2014, 4, 30, 15, 0, tzinfo=tzinfo),
            ],
        )

        # Check for number_of_output_files=4
        application.config.number_of_output_files = 4
        dates = []
        for d in application._dates_to_calculate:
            dates.append(d)
        self.assertEqual(
            dates,
            [
                dt.datetime(2014, 4, 30, 12, 0, tzinfo=tzinfo),
                dt.datetime(2014, 4, 30, 13, 0, tzinfo=tzinfo),
                dt.datetime(2014, 4, 30, 14, 0, tzinfo=tzinfo),
                dt.datetime(2014, 4, 30, 15, 0, tzinfo=tzinfo),
            ],
        )

    def test_dates_to_calculate_error1(self):
        self._create_mask_file()
        application = cli.App(self.config_file)
        with open(self.filenames[0], "a") as f:
            f.write(
                textwrap.dedent(
                    """\
                    2014-04-30 16:00,23.4,
                    malformed date,24.4,
                    2014-04-30 18:00,25.4,
                    2014-04-30 19:00,25.4,
                    """
                )
            )
        msg = (
            r"^Unable to parse date string u?'malformed date' "
            r"\(file " + self.filenames[0] + r", 3 lines from the end\)$"
        )
        with self.assertRaisesRegex(click.ClickException, msg):
            application.run()

    @patch("hspatial.cli.App._execute")
    def test_date_fmt(self, m):
        application = cli.App(self.config_file)
        application.run()

        # Ten-minute
        with open(self.filenames[0], "w") as f:
            f.write("Time_step=10min\n\n")
        with open(self.filenames[1], "w") as f:
            f.write("Time_step=10min\n\n")
        self.assertEqual(application._date_fmt, "%Y-%m-%d %H:%M%z")

        # Hourly
        with open(self.filenames[0], "w") as f:
            f.write("Time_step=h\n\n")
        with open(self.filenames[1], "w") as f:
            f.write("Time_step=h\n\n")
        self.assertEqual(application._date_fmt, "%Y-%m-%d %H:%M%z")

        # Daily
        with open(self.filenames[0], "w") as f:
            f.write("Time_step=D\n\n")
        with open(self.filenames[1], "w") as f:
            f.write("Time_step=D\n\n")
        self.assertEqual(application._date_fmt, "%Y-%m-%d")

        # Monthly
        with open(self.filenames[0], "w") as f:
            f.write("Time_step=M\n\n")
        with open(self.filenames[1], "w") as f:
            f.write("Time_step=M\n\n")
        self.assertEqual(application._date_fmt, "%Y-%m")

        # Annual
        with open(self.filenames[0], "w") as f:
            f.write("Time_step=Y\n\n")
        with open(self.filenames[1], "w") as f:
            f.write("Time_step=Y\n\n")
        self.assertEqual(application._date_fmt, "%Y")

        # Inconsistent
        with open(self.filenames[0], "w") as f:
            f.write("Time_step=10min\n\n")
        with open(self.filenames[1], "w") as f:
            f.write("Time_step=60min\n\n")
        with self.assertRaises(click.ClickException):
            application._date_fmt

    @patch("hspatial.cli.App._execute")
    def test_delete_obsolete_files(self, m):
        application = cli.App(self.config_file)
        application.run()

        # Create three files
        prefix = application.config.filename_prefix
        filename1 = os.path.join(self.output_dir, "{}-1.tif".format(prefix))
        filename2 = os.path.join(self.output_dir, "{}-2.tif".format(prefix))
        filename3 = os.path.join(self.output_dir, "{}-3.tif".format(prefix))
        Path(filename1).touch()
        Path(filename2).touch()
        Path(filename3).touch()

        # Just to make sure we didn't screw up above, check
        self.assertTrue(os.path.exists(filename1))
        self.assertTrue(os.path.exists(filename2))
        self.assertTrue(os.path.exists(filename3))

        # Execute for number_of_output_files = 2 and check
        application.config.number_of_output_files = 2
        application._delete_obsolete_files()
        self.assertFalse(os.path.exists(filename1))
        self.assertTrue(os.path.exists(filename2))
        self.assertTrue(os.path.exists(filename3))

        # Re-execute; nothing should have changed
        application._delete_obsolete_files()
        self.assertFalse(os.path.exists(filename1))
        self.assertTrue(os.path.exists(filename2))
        self.assertTrue(os.path.exists(filename3))

    def test_run(self):
        application = cli.App(self.config_file)

        # Create a mask
        self._create_mask_file()

        # Execute
        self._prepare_config_file(number_of_output_files=3)
        application.run()

        # Check that it has created three files
        full_prefix = os.path.join(self.output_dir, application.config.filename_prefix)
        self.assertTrue(os.path.exists(full_prefix + "-2014-04-30-15-00+0200.tif"))
        self.assertTrue(os.path.exists(full_prefix + "-2014-04-30-14-00+0200.tif"))
        self.assertTrue(os.path.exists(full_prefix + "-2014-04-30-13-00+0200.tif"))

        # Check the timestamp in the last file
        fp = gdal.Open(full_prefix + "-2014-04-30-15-00+0200.tif")
        timestamp = fp.GetMetadata()["TIMESTAMP"]
        fp = None
        self.assertEqual(timestamp, "2014-04-30 15:00+0200")

        # We could check a myriad other things here, but since we've
        # unit-tested lower level functions in detail, the above is reasonably
        # sufficient for us to know that it works.

    def test_no_timezone(self):
        self._remove_timezone_from_file(self.filenames[0])
        self._remove_timezone_from_file(self.filenames[1])
        application = cli.App(self.config_file)
        self._create_mask_file()
        self._prepare_config_file(number_of_output_files=3)
        msg = "{} does not contain Timezone".format(self.filenames[0])
        with self.assertRaisesRegex(click.ClickException, msg):
            application.run()

    def _remove_timezone_from_file(self, filename):
        with open(filename, "r") as f:
            lines = f.readlines()
        with open(filename, "w") as f:
            for line in lines:
                if not line.startswith("Timezone="):
                    f.write(line)



import datetime as dt
import re
import textwrap
from configparser import ParsingError
from copy import copy
from io import StringIO
from unittest import TestCase
from zoneinfo import ZoneInfo

import numpy as np
import pandas as pd
from iso8601 import parse_date

from htimeseries import FormatAutoDetector, HTimeseries, MetadataReader, MetadataWriter

tenmin_test_timeseries = textwrap.dedent(
    """\
    2008-02-07 11:20,1141.00,
    2008-02-07 11:30,1142.01,MISS
    2008-02-07 11:40,1154.02,
    2008-02-07 11:50,,
    2008-02-07 12:00,1180.04,
    """
)

tenmin_test_timeseries_file_version_2 = textwrap.dedent(
    """\
    Version=2\r
    Unit=°C\r
    Count=5\r
    Title=A test 10-min time series\r
    Comment=This timeseries is extremely important\r
    Comment=because the comment that describes it\r
    Comment=spans five lines.\r
    Comment=\r
    Comment=These five lines form two paragraphs.\r
    Timezone=+0200\r
    Time_step=10,0\r
    Variable=temperature\r
    Precision=1\r
    \r
    2008-02-07 11:20,1141.0,\r
    2008-02-07 11:30,1142.0,MISS\r
    2008-02-07 11:40,1154.0,\r
    2008-02-07 11:50,,\r
    2008-02-07 12:00,1180.0,\r
    """
)

tenmin_test_timeseries_file_version_3 = textwrap.dedent(
    """\
    Unit=°C\r
    Count=5\r
    Title=A test 10-min time series\r
    Comment=This timeseries is extremely important\r
    Comment=because the comment that describes it\r
    Comment=spans five lines.\r
    Comment=\r
    Comment=These five lines form two paragraphs.\r
    Timezone=+0200\r
    Time_step=10,0\r
    Variable=temperature\r
    Precision=1\r
    Location=24.678900 38.123450 4326\r
    Altitude=219.22\r
    \r
    2008-02-07 11:20,1141.0,\r
    2008-02-07 11:30,1142.0,MISS\r
    2008-02-07 11:40,1154.0,\r
    2008-02-07 11:50,,\r
    2008-02-07 12:00,1180.0,\r
    """
)

tenmin_test_timeseries_file_version_4 = textwrap.dedent(
    """\
    Unit=°C\r
    Count=5\r
    Title=A test 10-min time series\r
    Comment=This timeseries is extremely important\r
    Comment=because the comment that describes it\r
    Comment=spans five lines.\r
    Comment=\r
    Comment=These five lines form two paragraphs.\r
    Timezone=+0200\r
    Time_step=10,0\r
    Variable=temperature\r
    Precision=1\r
    Location=24.678900 38.123450 4326\r
    Altitude=219.22\r
    \r
    2008-02-07 11:20,1141.0,\r
    2008-02-07 11:30,1142.0,MISS\r
    2008-02-07 11:40,1154.0,\r
    2008-02-07 11:50,,\r
    2008-02-07 12:00,1180.0,\r
    """
)

tenmin_test_timeseries_file_no_altitude = textwrap.dedent(
    """\
    Unit=°C\r
    Count=5\r
    Title=A test 10-min time series\r
    Comment=This timeseries is extremely important\r
    Comment=because the comment that describes it\r
    Comment=spans five lines.\r
    Comment=\r
    Comment=These five lines form two paragraphs.\r
    Timezone=+0200\r
    Time_step=10,0\r
    Variable=temperature\r
    Precision=1\r
    Location=24.678900 38.123450 4326\r
    \r
    2008-02-07 11:20,1141.0,\r
    2008-02-07 11:30,1142.0,MISS\r
    2008-02-07 11:40,1154.0,\r
    2008-02-07 11:50,,\r
    2008-02-07 12:00,1180.0,\r
    """
)

tenmin_test_timeseries_file_no_location = textwrap.dedent(
    """\
    Unit=°C\r
    Count=5\r
    Title=A test 10-min time series\r
    Comment=This timeseries is extremely important\r
    Comment=because the comment that describes it\r
    Comment=spans five lines.\r
    Comment=\r
    Comment=These five lines form two paragraphs.\r
    Timezone=+0200\r
    Time_step=10,0\r
    Variable=temperature\r
    Precision=1\r
    \r
    2008-02-07 11:20,1141.0,\r
    2008-02-07 11:30,1142.0,MISS\r
    2008-02-07 11:40,1154.0,\r
    2008-02-07 11:50,,\r
    2008-02-07 12:00,1180.0,\r
    """
)

tenmin_test_timeseries_file_no_precision = textwrap.dedent(
    """\
    Unit=°C\r
    Count=5\r
    Title=A test 10-min time series\r
    Comment=This timeseries is extremely important\r
    Comment=because the comment that describes it\r
    Comment=spans five lines.\r
    Comment=\r
    Comment=These five lines form two paragraphs.\r
    Timezone=+0200\r
    Time_step=10,0\r
    Variable=temperature\r
    Location=24.678900 38.123450 4326\r
    Altitude=219.22\r
    \r
    2008-02-07 11:20,1141.000000,\r
    2008-02-07 11:30,1142.010000,MISS\r
    2008-02-07 11:40,1154.020000,\r
    2008-02-07 11:50,,\r
    2008-02-07 12:00,1180.040000,\r
    """
)

tenmin_test_timeseries_file_zero_precision = textwrap.dedent(
    """\
    Unit=°C\r
    Count=5\r
    Title=A test 10-min time series\r
    Comment=This timeseries is extremely important\r
    Comment=because the comment that describes it\r
    Comment=spans five lines.\r
    Comment=\r
    Comment=These five lines form two paragraphs.\r
    Timezone=+0200\r
    Time_step=10,0\r
    Variable=temperature\r
    Precision=0\r
    Location=24.678900 38.123450 4326\r
    Altitude=219.22\r
    \r
    2008-02-07 11:20,1141,\r
    2008-02-07 11:30,1142,MISS\r
    2008-02-07 11:40,1154,\r
    2008-02-07 11:50,,\r
    2008-02-07 12:00,1180,\r
    """
)


tenmin_test_timeseries_file_negative_precision = textwrap.dedent(
    """\
    Unit=°C\r
    Count=5\r
    Title=A test 10-min time series\r
    Comment=This timeseries is extremely important\r
    Comment=because the comment that describes it\r
    Comment=spans five lines.\r
    Comment=\r
    Comment=These five lines form two paragraphs.\r
    Timezone=+0200\r
    Time_step=10,0\r
    Variable=temperature\r
    Precision=-1\r
    Location=24.678900 38.123450 4326\r
    Altitude=219.22\r
    \r
    2008-02-07 11:20,1140,\r
    2008-02-07 11:30,1140,MISS\r
    2008-02-07 11:40,1150,\r
    2008-02-07 11:50,,\r
    2008-02-07 12:00,1180,\r
    """
)

standard_empty_dataframe = pd.DataFrame(
    data={"value": np.array([], dtype=np.float64), "flags": np.array([], dtype=str)},
    index=pd.DatetimeIndex([], tz=dt.timezone(dt.timedelta(hours=2))),
    columns=["value", "flags"],
)
standard_empty_dataframe.index.name = "date"


class HTimeseriesArgumentsTestCase(TestCase):
    def test_raises_on_invalid_argument(self):
        msg = r"HTimeseries.__init__\(\) got an unexpected keyword argument 'invalid'"
        with self.assertRaisesRegex(TypeError, msg):
            HTimeseries(invalid=42)

    def test_raises_if_timezone_unspecified(self):
        with self.assertRaises(TypeError):
            HTimeseries(StringIO(tenmin_test_timeseries))

    def test_raises_if_dataframe_naive(self):
        df = copy(standard_empty_dataframe)
        df.index = pd.DatetimeIndex([])  # Replace with a naive index
        with self.assertRaises(TypeError):
            HTimeseries(df)


class HTimeseriesEmptyTestCase(TestCase):
    def test_read_empty(self):
        s = StringIO()
        ts = HTimeseries(s, default_tzinfo=dt.timezone(dt.timedelta(hours=2)))
        pd.testing.assert_frame_equal(ts.data, standard_empty_dataframe)

    def test_write_empty(self):
        ts = HTimeseries(default_tzinfo=dt.timezone(dt.timedelta(hours=2)))
        s = StringIO()
        ts.write(s)
        self.assertEqual(s.getvalue(), "")

    def test_create_empty(self):
        pd.testing.assert_frame_equal(
            HTimeseries(default_tzinfo=dt.timezone(dt.timedelta(hours=2))).data,
            standard_empty_dataframe,
        )

    def test_unspecified_default_tzinfo(self):
        ts = HTimeseries()
        self.assertEqual(ts.data.index.tz, dt.timezone.utc)


class HTimeseriesWriteSimpleTestCase(TestCase):
    def test_write(self):
        anp = np.array(
            [
                [parse_date("2005-08-23 18:53"), 93, ""],
                [parse_date("2005-08-24 19:52"), 108.7, ""],
                [parse_date("2005-08-25 23:59"), 28.3, "HEARTS SPADES"],
                [parse_date("2005-08-26 00:02"), float("NaN"), ""],
                [parse_date("2005-08-27 00:02"), float("NaN"), "DIAMONDS"],
            ]
        )
        data = pd.DataFrame(anp[:, [1, 2]], index=anp[:, 0], columns=("value", "flags"))
        ts = HTimeseries(data=data)
        s = StringIO()
        ts.write(s)
        self.assertEqual(
            s.getvalue(),
            textwrap.dedent(
                """\
                2005-08-23 18:53,93,\r
                2005-08-24 19:52,108.7,\r
                2005-08-25 23:59,28.3,HEARTS SPADES\r
                2005-08-26 00:02,,\r
                2005-08-27 00:02,,DIAMONDS\r
                """
            ),
        )


class HTimeseriesWriteFileTestCase(TestCase):
    def setUp(self):
        data = pd.read_csv(
            StringIO(tenmin_test_timeseries),
            parse_dates=[0],
            usecols=["date", "value", "flags"],
            index_col=0,
            header=None,
            names=("date", "value", "flags"),
            dtype={"value": np.float64, "flags": str},
        ).asfreq("10min")
        data.index = data.index.tz_localize(dt.timezone(dt.timedelta(hours=2)))
        self.reference_ts = HTimeseries(data=data)
        self.reference_ts.unit = "°C"
        self.reference_ts.title = "A test 10-min time series"
        self.reference_ts.precision = 1
        self.reference_ts.time_step = "10min"
        self.reference_ts.variable = "temperature"
        self.reference_ts.comment = (
            "This timeseries is extremely important\n"
            "because the comment that describes it\n"
            "spans five lines.\n\n"
            "These five lines form two paragraphs."
        )
        self.reference_ts.location = {
            "abscissa": 24.6789,
            "ordinate": 38.12345,
            "srid": 4326,
            "altitude": 219.22,
            "asrid": None,
        }

    def test_version_2(self):
        outstring = StringIO()
        self.reference_ts.write(outstring, format=HTimeseries.FILE, version=2)
        self.assertEqual(outstring.getvalue(), tenmin_test_timeseries_file_version_2)

    def test_version_3(self):
        outstring = StringIO()
        self.reference_ts.write(outstring, format=HTimeseries.FILE, version=3)
        self.assertEqual(outstring.getvalue(), tenmin_test_timeseries_file_version_3)

    def test_version_4(self):
        outstring = StringIO()
        self.reference_ts.write(outstring, format=HTimeseries.FILE, version=4)
        self.assertEqual(outstring.getvalue(), tenmin_test_timeseries_file_version_4)

    def test_version_5(self):
        outstring = StringIO()
        self.reference_ts.write(outstring, format=HTimeseries.FILE, version=5)
        self.assertEqual(
            outstring.getvalue(),
            tenmin_test_timeseries_file_version_4.replace(
                "Time_step=10,0", "Time_step=10min"
            ),
        )

    def test_version_latest(self):
        outstring = StringIO()
        self.reference_ts.write(outstring, format=HTimeseries.FILE)
        self.assertEqual(
            outstring.getvalue(),
            tenmin_test_timeseries_file_version_4.replace(
                "Time_step=10,0", "Time_step=10min"
            ),
        )

    def test_altitude_none(self):
        self.reference_ts.location["altitude"] = None
        outstring = StringIO()
        self.reference_ts.write(outstring, format=HTimeseries.FILE, version=4)
        self.assertEqual(outstring.getvalue(), tenmin_test_timeseries_file_no_altitude)

    def test_no_altitude(self):
        del self.reference_ts.location["altitude"]
        outstring = StringIO()
        self.reference_ts.write(outstring, format=HTimeseries.FILE, version=4)
        self.assertEqual(outstring.getvalue(), tenmin_test_timeseries_file_no_altitude)

    def test_altitude_zero(self):
        self.reference_ts.location["altitude"] = 0
        outstring = StringIO()
        self.reference_ts.write(outstring, format=HTimeseries.FILE, version=4)
        self.assertIn("Altitude=0", outstring.getvalue())

    def test_location_none(self):
        self.reference_ts.location = None
        outstring = StringIO()
        self.reference_ts.write(outstring, format=HTimeseries.FILE, version=4)
        self.assertEqual(outstring.getvalue(), tenmin_test_timeseries_file_no_location)

    def test_no_location(self):
        delattr(self.reference_ts, "location")
        outstring = StringIO()
        self.reference_ts.write(outstring, format=HTimeseries.FILE, version=4)
        self.assertEqual(outstring.getvalue(), tenmin_test_timeseries_file_no_location)

    def test_precision_none(self):
        self.reference_ts.precision = None
        outstring = StringIO()
        self.reference_ts.write(outstring, format=HTimeseries.FILE, version=4)
        self.assertEqual(outstring.getvalue(), tenmin_test_timeseries_file_no_precision)

    def test_no_precision(self):
        delattr(self.reference_ts, "precision")
        outstring = StringIO()
        self.reference_ts.write(outstring, format=HTimeseries.FILE, version=4)
        self.assertEqual(outstring.getvalue(), tenmin_test_timeseries_file_no_precision)

    def test_precision_zero(self):
        self.reference_ts.precision = 0
        outstring = StringIO()
        self.reference_ts.write(outstring, format=HTimeseries.FILE, version=4)
        self.assertEqual(
            outstring.getvalue(), tenmin_test_timeseries_file_zero_precision
        )

    def test_negative_precision(self):
        self.reference_ts.precision = -1
        outstring = StringIO()
        self.reference_ts.write(outstring, format=HTimeseries.FILE, version=4)
        self.assertEqual(
            outstring.getvalue(), tenmin_test_timeseries_file_negative_precision
        )

    def test_timezone_utc(self):
        self.reference_ts.data = self.reference_ts.data.tz_convert(dt.timezone.utc)
        outstring = StringIO()
        self.reference_ts.write(outstring, format=HTimeseries.FILE, version=4)
        self.assertIn("Timezone=+0000\r\n", outstring.getvalue())

    def test_timezone_positive(self):
        tz = dt.timezone(dt.timedelta(hours=2, minutes=30))
        self.reference_ts.data = self.reference_ts.data.tz_convert(tz)
        outstring = StringIO()
        self.reference_ts.write(outstring, format=HTimeseries.FILE, version=4)
        self.assertIn("Timezone=+0230\r\n", outstring.getvalue())

    def test_timezone_negative(self):
        tz = dt.timezone(-dt.timedelta(hours=3, minutes=15))
        self.reference_ts.data = self.reference_ts.data.tz_convert(tz)
        outstring = StringIO()
        self.reference_ts.write(outstring, format=HTimeseries.FILE, version=4)
        self.assertIn("Timezone=-0315\r\n", outstring.getvalue())


class ReadFilelikeTestCaseBase:
    def test_length(self):
        self.assertEqual(len(self.ts.data), 5)

    def test_dates(self):
        np.testing.assert_array_equal(
            self.ts.data.index,
            pd.date_range("2008-02-07 11:20+0200", periods=5, freq="10min"),
        )

    def test_values(self):
        expected = np.array(
            [1141.00, 1142.01, 1154.02, float("NaN"), 1180.04], dtype=float
        )
        np.testing.assert_allclose(self.ts.data.values[:, 0].astype(float), expected)

    def test_flags(self):
        expected = np.array(["", "MISS", "", "", ""])
        np.testing.assert_array_equal(self.ts.data.values[:, 1], expected)

    def test_tz(self):
        self.assertEqual(
            self.ts.data.index.tz.utcoffset(dt.datetime(2000, 1, 1)),
            dt.timedelta(hours=2),
        )


class HTimeseriesReadTwoColumnsTestCase(ReadFilelikeTestCaseBase, TestCase):
    def setUp(self):
        string = self._remove_flags_column(tenmin_test_timeseries)
        s = StringIO(string)
        s.seek(0)
        self.ts = HTimeseries(s, default_tzinfo=ZoneInfo("Etc/GMT-2"))

    def _remove_flags_column(self, s):
        return re.sub(r",[^,]*$", "", s, flags=re.MULTILINE) + "\n"

    def test_flags(self):
        expected = np.array(["", "", "", "", ""])
        np.testing.assert_array_equal(self.ts.data.values[:, 1], expected)

    def test_tz(self):
        self.assertEqual(
            self.ts.data.index.tz.utcoffset(dt.datetime(2000, 1, 1)),
            dt.timedelta(hours=2),
        )


class HTimeseriesReadMixOf2And3ColumnsTestCase(ReadFilelikeTestCaseBase, TestCase):
    def setUp(self):
        string = self._remove_empty_flags_column(tenmin_test_timeseries)
        s = StringIO(string)
        s.seek(0)
        self.ts = HTimeseries(s, default_tzinfo=ZoneInfo("Etc/GMT-2"))

    def _remove_empty_flags_column(self, s):
        return re.sub(r",$", "", s, flags=re.MULTILINE) + "\n"


class HTimeseriesReadOneColumnTestCase(TestCase):
    def test_one_column(self):
        s = StringIO("2023-12-19 15:17\n")
        s.seek(0)
        self.ts = HTimeseries(s, default_tzinfo=ZoneInfo("Etc/GMT-2"))
        expected = pd.DataFrame(
            {"value": [np.nan], "flags": [""]},
            index=[dt.datetime(2023, 12, 19, 15, 17, tzinfo=ZoneInfo("Etc/GMT-2"))],
        )
        expected.index.name = "date"
        pd.testing.assert_frame_equal(self.ts.data, expected)


class HTimeseriesReadFilelikeMetadataOnlyTestCase(TestCase):
    def setUp(self):
        s = StringIO(tenmin_test_timeseries_file_version_4)
        s.seek(0)
        self.ts = HTimeseries(
            s, start_date="1971-01-01 00:00", end_date="1970-01-01 00:00"
        )

    def test_data_is_empty(self):
        self.assertEqual(len(self.ts.data), 0)

    def test_metadata_was_read(self):
        self.assertEqual(self.ts.unit, "°C")


class HTimeseriesReadFilelikeWithMissingLocationButPresentAltitudeTestCase(TestCase):
    def setUp(self):
        s = StringIO("Altitude=55\n\n")
        self.ts = HTimeseries(s, default_tzinfo=dt.timezone(dt.timedelta(hours=2)))

    def test_data_is_empty(self):
        pd.testing.assert_frame_equal(self.ts.data, standard_empty_dataframe)

    def test_has_altitude(self):
        self.assertEqual(self.ts.location["altitude"], 55)

    def test_has_no_abscissa(self):
        self.assertFalse("abscissa" in self.ts.location)


class HTimeseriesReadWithStartDateAndEndDateTestCase(TestCase):
    def setUp(self):
        s = StringIO(tenmin_test_timeseries)
        s.seek(0)
        self.ts = HTimeseries(
            s,
            start_date=dt.datetime(2008, 2, 7, 11, 30, tzinfo=dt.timezone.utc),
            end_date=dt.datetime(2008, 2, 7, 11, 55, tzinfo=dt.timezone.utc),
            default_tzinfo=dt.timezone.utc,
        )

    def test_length(self):
        self.assertEqual(len(self.ts.data), 3)

    def test_dates(self):
        np.testing.assert_array_equal(
            self.ts.data.index,
            pd.date_range(
                "2008-02-07 11:30", periods=3, freq="10min", tz=dt.timezone.utc
            ),
        )

    def test_values(self):
        np.testing.assert_allclose(
            self.ts.data.values[:, 0].astype(float),
            np.array([1142.01, 1154.02, float("NaN")]),
        )

    def test_flags(self):
        np.testing.assert_array_equal(
            self.ts.data.values[:, 1], np.array(["MISS", "", ""])
        )


class HTimeseriesReadWithStartDateTestCase(TestCase):
    def setUp(self):
        s = StringIO(tenmin_test_timeseries)
        s.seek(0)
        self.ts = HTimeseries(
            s,
            start_date=dt.datetime(2008, 2, 7, 11, 45, tzinfo=dt.timezone.utc),
            default_tzinfo=dt.timezone.utc,
        )

    def test_length(self):
        self.assertEqual(len(self.ts.data), 2)

    def test_dates(self):
        np.testing.assert_array_equal(
            self.ts.data.index,
            pd.date_range(
                "2008-02-07 11:50", periods=2, freq="10min", tz=dt.timezone.utc
            ),
        )

    def test_values(self):
        np.testing.assert_allclose(
            self.ts.data.values[:, 0].astype(float), np.array([float("NaN"), 1180.04])
        )

    def test_flags(self):
        np.testing.assert_array_equal(self.ts.data.values[:, 1], np.array(["", ""]))


class HTimeseriesReadWithEndDateTestCase(TestCase):
    def setUp(self):
        s = StringIO(tenmin_test_timeseries)
        s.seek(0)
        self.ts = HTimeseries(
            s,
            end_date=dt.datetime(2008, 2, 7, 11, 50, tzinfo=dt.timezone.utc),
            default_tzinfo=dt.timezone.utc,
        )

    def test_length(self):
        self.assertEqual(len(self.ts.data), 4)

    def test_dates(self):
        np.testing.assert_array_equal(
            self.ts.data.index,
            pd.date_range(
                "2008-02-07 11:20", periods=4, freq="10min", tz=dt.timezone.utc
            ),
        )

    def test_values(self):
        np.testing.assert_allclose(
            self.ts.data.values[:, 0].astype(float),
            np.array([1141.00, 1142.01, 1154.02, float("NaN")]),
        )

    def test_flags(self):
        np.testing.assert_array_equal(
            self.ts.data.values[:, 1], np.array(["", "MISS", "", ""])
        )


class HTimeseriesReadFileFormatTestCase(TestCase):
    def setUp(self):
        s = StringIO(tenmin_test_timeseries_file_version_4)
        s.seek(0)
        self.ts = HTimeseries(s)

    def test_unit(self):
        self.assertEqual(self.ts.unit, "°C")

    def test_title(self):
        self.assertEqual(self.ts.title, "A test 10-min time series")

    def test_comment(self):
        self.assertEqual(
            self.ts.comment,
            textwrap.dedent(
                """\
                This timeseries is extremely important
                because the comment that describes it
                spans five lines.

                These five lines form two paragraphs."""
            ),
        )

    def test_timezone(self):
        self.assertEqual(self.ts.data.index.tz.utcoffset(None), dt.timedelta(hours=2))

    def test_time_step(self):
        self.assertEqual(self.ts.time_step, "10min")

    def test_variable(self):
        self.assertEqual(self.ts.variable, "temperature")

    def test_precision(self):
        self.assertEqual(self.ts.precision, 1)

    def test_abscissa(self):
        self.assertAlmostEqual(self.ts.location["abscissa"], 24.678900, places=6)

    def test_ordinate(self):
        self.assertAlmostEqual(self.ts.location["ordinate"], 38.123450, places=6)

    def test_srid(self):
        self.assertEqual(self.ts.location["srid"], 4326)

    def test_altitude(self):
        self.assertAlmostEqual(self.ts.location["altitude"], 219.22, places=2)

    def test_asrid(self):
        self.assertTrue(self.ts.location["asrid"] is None)

    def test_length(self):
        self.assertEqual(len(self.ts.data), 5)

    def test_dates(self):
        np.testing.assert_array_equal(
            self.ts.data.index,
            pd.date_range(
                "2008-02-07 11:20", periods=5, freq="10min", tz=ZoneInfo("Etc/GMT-2")
            ),
        )

    def test_values(self):
        np.testing.assert_allclose(
            self.ts.data.values[:, 0].astype(float),
            np.array([1141.0, 1142.0, 1154.0, float("NaN"), 1180.0]),
        )

    def test_flags(self):
        np.testing.assert_array_equal(
            self.ts.data.values[:, 1], np.array(["", "MISS", "", "", ""])
        )


class FormatAutoDetectorTestCase(TestCase):
    def test_auto_detect_text_format(self):
        detected_format = FormatAutoDetector(StringIO(tenmin_test_timeseries)).detect()
        self.assertEqual(detected_format, HTimeseries.TEXT)

    def test_auto_detect_file_format(self):
        detected_format = FormatAutoDetector(
            StringIO(tenmin_test_timeseries_file_version_4)
        ).detect()
        self.assertEqual(detected_format, HTimeseries.FILE)


class WriteOldTimeStepTestCase(TestCase):
    def get_value(self, time_step):
        self.f = StringIO()
        self.htimeseries = HTimeseries()
        self.htimeseries.time_step = time_step
        MetadataWriter(self.f, self.htimeseries, version=2).write_time_step()
        return self.f.getvalue()

    def test_empty(self):
        self.assertEqual(self.get_value(""), "")

    def test_min(self):
        self.assertEqual(self.get_value("27min"), "Time_step=27,0\r\n")

    def test_min_without_number(self):
        self.assertEqual(self.get_value("min"), "Time_step=1,0\r\n")

    def test_hour(self):
        self.assertEqual(self.get_value("3h"), "Time_step=180,0\r\n")

    def test_hour_without_number(self):
        self.assertEqual(self.get_value("h"), "Time_step=60,0\r\n")

    def test_day(self):
        self.assertEqual(self.get_value("3D"), "Time_step=4320,0\r\n")

    def test_day_without_number(self):
        self.assertEqual(self.get_value("D"), "Time_step=1440,0\r\n")

    def test_month(self):
        self.assertEqual(self.get_value("3ME"), "Time_step=0,3\r\n")

    def test_month_without_number(self):
        self.assertEqual(self.get_value("ME"), "Time_step=0,1\r\n")

    def test_year(self):
        self.assertEqual(self.get_value("3YE"), "Time_step=0,36\r\n")

    def test_year_without_number(self):
        self.assertEqual(self.get_value("YE"), "Time_step=0,12\r\n")

    def test_garbage(self):
        with self.assertRaisesRegex(ValueError, 'Cannot format time step "hello"'):
            self.get_value("hello")

    def test_wrong_number(self):
        with self.assertRaisesRegex(ValueError, 'Cannot format time step "FM"'):
            self.get_value("FM")


class GetTimeStepTestCase(TestCase):
    def get_value(self, time_step):
        f = StringIO("Time_step={}\r\n\r\n".format(time_step))
        return MetadataReader(f).meta["time_step"]

    def test_min(self):
        self.assertEqual(self.get_value("1min"), "1min")

    def test_minutes(self):
        self.assertEqual(self.get_value("250,0"), "250min")

    def test_months(self):
        self.assertEqual(self.get_value("0,25"), "25M")

    def test_both_nonzero(self):
        with self.assertRaisesRegex(ParsingError, "Invalid time step"):
            self.get_value("5,5")


class HTimeseriesReadWithDuplicateDatesTestCase(TestCase):
    csv_with_duplicates = textwrap.dedent(
        """\
        2020-02-23 11:00,5,
        2020-02-23 12:00,6,
        2020-02-23 12:00,7,
        2020-02-23 13:00,8,
        2020-02-23 13:00,8,
        2020-02-23 14:00,8,
        """
    )

    def test_read_csv_with_duplicates_raises_error(self):
        s = StringIO(self.csv_with_duplicates)
        s.seek(0)
        msg = (
            "Can't read time series: the following timestamps appear more than once: "
            "2020-02-23 12:00:00, 2020-02-23 13:00:00"
        )
        with self.assertRaisesRegex(ValueError, msg):
            HTimeseries(s)

    def test_write_csv_with_duplicates_raises_error(self):
        data = pd.read_csv(
            StringIO(self.csv_with_duplicates),
            parse_dates=[0],
            usecols=["date", "value", "flags"],
            index_col=0,
            header=None,
            names=("date", "value", "flags"),
            dtype={"value": np.float64, "flags": str},
        )
        data.index = data.index.tz_localize(dt.timezone.utc)
        msg = (
            "Can't write time series: the following timestamps appear more than once: "
            r"2020-02-23 12:00:00\+00:00, 2020-02-23 13:00:00\+00:00"
        )
        with self.assertRaisesRegex(ValueError, msg):
            HTimeseries(data).write(StringIO())


class HTimeseriesTimeChangeTestCase(TestCase):
    """Test what happens when we read a csv containing a time change.

    We use a hard case here, a switch from DST to normal, which contains a duplicate
    hour. HTimeseries will refuse to handle repeating timestamps, so we use test data
    that does not contain a repeating hour. In that case, HTimeseries assumes the
    ambiguous times are before the switch.
    """

    time_change_test_timeseries = textwrap.dedent(
        """\
        2023-10-29 02:30,15,
        2023-10-29 03:00,16,
        2023-10-29 03:30,17,
        2023-10-29 04:00,20,
        2023-10-29 04:30,21,
        """
    )

    def setUp(self):
        s = StringIO(self.time_change_test_timeseries)
        s.seek(0)
        self.ts = HTimeseries(s, default_tzinfo=ZoneInfo("Europe/Athens"))

    def test_dates(self):
        expected = np.array(
            [
                dt.datetime(2023, 10, 28, 23, 30, 0, tzinfo=dt.timezone.utc),
                dt.datetime(2023, 10, 29, 0, 0, 0, tzinfo=dt.timezone.utc),
                dt.datetime(2023, 10, 29, 0, 30, 0, tzinfo=dt.timezone.utc),
                dt.datetime(2023, 10, 29, 2, 0, 0, tzinfo=dt.timezone.utc),
                dt.datetime(2023, 10, 29, 2, 30, 0, tzinfo=dt.timezone.utc),
            ]
        )
        np.testing.assert_array_equal(
            self.ts.data.index.tz_convert(dt.timezone.utc),
            expected,
        )


class HTimeseriesCsvWithAwareTimestampsTestCase(TestCase):
    """Test what happens when we read a csv with aware timestamps.

    Some enhydris-api-client/loggertodb versions attempt to upload CSVs
    with timestamps containing time zones, such as 2023-12-22
    19:53+00:00. We check that we handle this correctly.
    """

    def test_csv_with_aware_timestamps(self):
        s = StringIO("2023-12-22 19:53+00:00,42.0,\n")
        s.seek(0)
        self.ts = HTimeseries(s, default_tzinfo=ZoneInfo("Etc/GMT-2"))
        self.assertEqual(
            self.ts.data.index[0],
            dt.datetime(2023, 12, 22, 19, 53, tzinfo=dt.timezone.utc),
        )

    def test_csv_with_mixed_timestamps(self):
        s = StringIO("2023-12-22 19:53+00:00,42.0,\n2023-12-22 20:53,43.0,\n")
        s.seek(0)
        msg = "Maybe the CSV contains mixed aware and naive timestamps"
        with self.assertRaisesRegex(ValueError, msg):
            HTimeseries(s, default_tzinfo=ZoneInfo("Etc/GMT-2"))



import datetime as dt
import math
from unittest import TestCase

import numpy as np

from evaporation import PenmanMonteith


class SenegalTzinfo(dt.tzinfo):
    """
    At various places we test using Example 19, p. 75, of Allen et al. (1998).
    The example calculates evaporation in Senegal.  Although Senegal has time
    Afrika/Dakar, which is the same as UTC, Example 19 apparently assumes that
    its time zone is actually UTC-01:00 (which would be more consistent with
    its longitude, which may be the reason for the error).  So we make the same
    assumption as example 19, in order to get the same result.
    """

    def utcoffset(self, adate):
        return -dt.timedelta(hours=1)

    def dst(self, adate):
        return dt.timedelta(0)


senegal_tzinfo = SenegalTzinfo()


class PenmanMonteithTestCase(TestCase):
    def test_daily(self):
        # Apply Allen et al. (1998) Example 18 page 72.

        unit_converters = {
            # Eq. 47 p. 56
            "wind_speed": lambda x: x
            * 4.87
            / math.log(67.8 * 10 - 5.42)
        }

        pm = PenmanMonteith(
            albedo=0.23,
            elevation=100,
            latitude=50.8,
            time_step="D",
            unit_converters=unit_converters,
        )

        result = pm.calculate(
            temperature_max=21.5,
            temperature_min=12.3,
            humidity_max=84,
            humidity_min=63,
            wind_speed=2.78,
            sunshine_duration=9.25,
            adatetime=dt.date(2014, 7, 6),
        )
        self.assertAlmostEqual(result, 3.9, places=1)

        # We try the same calculation, but instead of sunshine duration we
        # provide the solar radiation directly. Should get the same result.
        result = pm.calculate(
            temperature_max=21.5,
            temperature_min=12.3,
            humidity_max=84,
            humidity_min=63,
            wind_speed=2.78,
            solar_radiation=22.07,
            adatetime=dt.date(2014, 7, 6),
        )
        self.assertAlmostEqual(result, 3.9, places=1)

    def test_daily_grid(self):
        # We use a 2x1 grid, where point 1, 1 is the same as Example 18, and
        # point 1, 2 has some different values.

        unit_converters = {
            # Eq. 47 p. 56
            "wind_speed": lambda x: x
            * 4.87
            / math.log(67.8 * 10 - 5.42)
        }

        pm = PenmanMonteith(
            albedo=0.23,
            elevation=100,
            latitude=50.8,
            time_step="D",
            unit_converters=unit_converters,
        )
        result = pm.calculate(
            temperature_max=np.array([21.5, 28]),
            temperature_min=np.array([12.3, 15]),
            humidity_max=np.array([84, 70]),
            humidity_min=np.array([63, 60]),
            wind_speed=np.array([2.78, 3]),
            sunshine_duration=np.array([9.25, 9]),
            adatetime=dt.date(2014, 7, 6),
        )
        np.testing.assert_almost_equal(result, np.array([3.9, 4.8]), decimal=1)

        # Same thing with solar radiation instead of sunshine duration
        result = pm.calculate(
            temperature_max=np.array([21.5, 28]),
            temperature_min=np.array([12.3, 15]),
            humidity_max=np.array([84, 70]),
            humidity_min=np.array([63, 60]),
            wind_speed=np.array([2.78, 3]),
            solar_radiation=np.array([22.07, 21.62]),
            adatetime=dt.date(2014, 7, 6),
        )
        np.testing.assert_almost_equal(result, np.array([3.9, 4.8]), decimal=1)

    def test_hourly(self):
        # Apply Allen et al. (1998) Example 19 page 75.
        pm = PenmanMonteith(
            albedo=0.23,
            nighttime_solar_radiation_ratio=0.8,
            elevation=8,
            latitude=16.217,
            longitude=-16.25,
            time_step="h",
        )

        result = pm.calculate(
            temperature=38,
            humidity=52,
            wind_speed=3.3,
            pressure=101.3,
            solar_radiation=2.450,
            adatetime=dt.datetime(2014, 10, 1, 15, 0, tzinfo=senegal_tzinfo),
        )
        self.assertAlmostEqual(result, 0.63, places=2)
        result = pm.calculate(
            temperature=28,
            humidity=90,
            wind_speed=1.9,
            pressure=101.3,
            solar_radiation=0,
            adatetime=dt.datetime(2014, 10, 1, 2, 30, tzinfo=senegal_tzinfo),
        )
        self.assertAlmostEqual(result, 0.0, places=2)

        # Same thing, but let it calculate pressure itself
        result = pm.calculate(
            temperature=38,
            humidity=52,
            wind_speed=3.3,
            solar_radiation=2.450,
            adatetime=dt.datetime(2014, 10, 1, 15, 0, tzinfo=senegal_tzinfo),
        )
        self.assertAlmostEqual(result, 0.63, places=2)
        result = pm.calculate(
            temperature=28,
            humidity=90,
            wind_speed=1.9,
            solar_radiation=0,
            adatetime=dt.datetime(2014, 10, 1, 2, 30, tzinfo=senegal_tzinfo),
        )
        self.assertAlmostEqual(result, 0.0, places=2)

    def test_hourly_grid(self):
        # We use a 2x1 grid, where point 1, 1 is the same as Example 19, and
        # point 1, 2 has some different values.
        pm = PenmanMonteith(
            albedo=0.23,
            nighttime_solar_radiation_ratio=0.8,
            elevation=8,
            latitude=16.217,
            longitude=np.array([-16.25, -15.25]),
            time_step="h",
        )
        result = pm.calculate(
            temperature=np.array([38, 28]),
            humidity=np.array([52, 42]),
            wind_speed=np.array([3.3, 2.3]),
            pressure=101.3,
            solar_radiation=np.array([2.450, 1.450]),
            adatetime=dt.datetime(2014, 10, 1, 15, 0, tzinfo=senegal_tzinfo),
        )
        np.testing.assert_almost_equal(result, np.array([0.63, 0.36]), decimal=2)

    def test_hourly_with_albedo_grid(self):
        # Apply Allen et al. (1998) Example 19 page 75.
        pm = PenmanMonteith(
            albedo=np.array([0.23]),
            nighttime_solar_radiation_ratio=0.8,
            elevation=8,
            latitude=16.217,
            longitude=-16.25,
            time_step="h",
        )

        result = pm.calculate(
            temperature=38,
            humidity=52,
            wind_speed=3.3,
            pressure=101.3,
            solar_radiation=2.450,
            adatetime=dt.datetime(2014, 10, 1, 15, 0, tzinfo=senegal_tzinfo),
        )
        # The following two lines could be written more simply like this:
        #     self.assertAlmostEqual(result, 0.63, places=2)
        # However, it does not work properly on Python 3 because of a numpy
        # issue.
        self.assertEqual(result.size, 1)
        self.assertAlmostEqual(result[0], 0.63, places=2)

    def test_hourly_array_with_seasonal_albedo_grid(self):
        # We use a 2x1 grid, where point 1, 1 is the same as Example 19, and
        # point 1, 2 has some different values.
        pm = PenmanMonteith(
            albedo=[np.array([0.23, 0.23]) for item in range(1, 13)],
            nighttime_solar_radiation_ratio=0.8,
            elevation=8,
            latitude=16.217,
            longitude=np.array([-16.25, -15.25]),
            time_step="h",
        )
        result = pm.calculate(
            temperature=np.array([38, 28]),
            humidity=np.array([52, 42]),
            wind_speed=np.array([3.3, 2.3]),
            pressure=101.3,
            solar_radiation=np.array([2.450, 1.450]),
            adatetime=dt.datetime(2014, 10, 1, 15, 0, tzinfo=senegal_tzinfo),
        )
        np.testing.assert_almost_equal(result, np.array([0.63, 0.36]), decimal=2)

    def test_hourly_with_seasonal_albedo(self):
        # Apply Allen et al. (1998) Example 19 page 75.

        pm = PenmanMonteith(
            albedo=[
                0.13,
                0.01,
                0.01,
                0.01,
                0.01,
                0.01,
                0.01,
                0.01,
                0.01,
                0.23,
                0.01,
                0.33,
            ],
            nighttime_solar_radiation_ratio=0.8,
            elevation=8,
            latitude=16.217,
            longitude=-16.25,
            time_step="h",
        )

        result = pm.calculate(
            temperature=38,
            humidity=52,
            wind_speed=3.3,
            pressure=101.3,
            solar_radiation=2.450,
            adatetime=dt.datetime(2014, 1, 1, 15, 0, tzinfo=senegal_tzinfo),
        )
        self.assertAlmostEqual(result, 0.69, places=2)

        result = pm.calculate(
            temperature=38,
            humidity=52,
            wind_speed=3.3,
            pressure=101.3,
            solar_radiation=2.450,
            adatetime=dt.datetime(2014, 12, 1, 15, 0, tzinfo=senegal_tzinfo),
        )
        self.assertAlmostEqual(result, 0.56, places=2)

        result = pm.calculate(
            temperature=38,
            humidity=52,
            wind_speed=3.3,
            pressure=101.3,
            solar_radiation=2.450,
            adatetime=dt.datetime(2014, 10, 1, 15, 0, tzinfo=senegal_tzinfo),
        )
        self.assertAlmostEqual(result, 0.63, places=2)



import datetime as dt
import math
import os
import shutil
import tempfile
import textwrap
from stat import S_IREAD, S_IRGRP, S_IROTH
from time import sleep
from unittest import TestCase

import numpy as np
import pandas as pd
from django.contrib.gis.gdal import GDALRaster
from django.contrib.gis.geos import Point as GeoDjangoPoint
from osgeo import gdal, ogr, osr

import hspatial
from hspatial.test import setup_test_raster
from htimeseries import HTimeseries, TzinfoFromString

gdal.UseExceptions()

UTC_PLUS_2 = dt.timezone(dt.timedelta(hours=2))


def add_point_to_layer(layer, x, y, value):
    p = ogr.Geometry(ogr.wkbPoint)
    p.AddPoint(x, y)
    f = ogr.Feature(layer.GetLayerDefn())
    f.SetGeometry(p)
    f.SetField("value", value)
    layer.CreateFeature(f)


class IdwTestCase(TestCase):
    def setUp(self):
        self.point = ogr.Geometry(ogr.wkbPoint)
        self.point.AddPoint(5.1, 2.5)

        self.data_source = ogr.GetDriverByName("memory").CreateDataSource("tmp")
        self.data_layer = self.data_source.CreateLayer("test")
        self.data_layer.CreateField(ogr.FieldDefn("value", ogr.OFTReal))

    def tearDown(self):
        self.data_layer = None
        self.data_source = None
        self.point = None

    def test_idw_single_point(self):
        add_point_to_layer(self.data_layer, 5.3, 6.4, 42.8)
        self.assertAlmostEqual(hspatial.idw(self.point, self.data_layer), 42.8)

    def test_idw_three_points(self):
        add_point_to_layer(self.data_layer, 6.4, 7.8, 33.0)
        add_point_to_layer(self.data_layer, 9.5, 7.4, 94.0)
        add_point_to_layer(self.data_layer, 7.1, 4.9, 67.7)
        self.assertAlmostEqual(
            hspatial.idw(self.point, self.data_layer), 64.090, places=3
        )
        self.assertAlmostEqual(
            hspatial.idw(self.point, self.data_layer, alpha=2.0), 64.188, places=3
        )

    def test_idw_point_with_nan(self):
        add_point_to_layer(self.data_layer, 6.4, 7.8, 33.0)
        add_point_to_layer(self.data_layer, 9.5, 7.4, 94.0)
        add_point_to_layer(self.data_layer, 7.1, 4.9, 67.7)
        add_point_to_layer(self.data_layer, 7.2, 5.4, float("nan"))
        self.assertAlmostEqual(
            hspatial.idw(self.point, self.data_layer), 64.090, places=3
        )
        self.assertAlmostEqual(
            hspatial.idw(self.point, self.data_layer, alpha=2.0), 64.188, places=3
        )


class IntegrateTestCase(TestCase):
    # The calculations for this test have been made manually in
    # data/spatial_calculations.ods, tab test_integrate_idw.

    def setUp(self):
        # We will test on a 7x15 grid
        self.mask = np.zeros((7, 15), np.int8)
        self.mask[3, 3] = 1
        self.mask[6, 14] = 1
        self.mask[4, 13] = 1
        self.dataset = gdal.GetDriverByName("mem").Create(
            "test", 15, 7, 1, gdal.GDT_Byte
        )
        self.dataset.GetRasterBand(1).WriteArray(self.mask)

        # Prepare the result band
        self.dataset.AddBand(gdal.GDT_Float32)
        self.target_band = self.dataset.GetRasterBand(self.dataset.RasterCount)

        # Our grid represents a 70x150m area, lower-left co-ordinates (0, 0).
        self.dataset.SetGeoTransform((0, 10, 0, 70, 0, -10))

        # Now the data layer, with three points
        self.data_source = ogr.GetDriverByName("memory").CreateDataSource("tmp")
        self.data_layer = self.data_source.CreateLayer("test")
        self.data_layer.CreateField(ogr.FieldDefn("value", ogr.OFTReal))
        add_point_to_layer(self.data_layer, 75.2, 10.7, 37.4)
        add_point_to_layer(self.data_layer, 125.7, 19.0, 24.0)
        add_point_to_layer(self.data_layer, 9.8, 57.1, 95.4)

    def tearDown(self):
        self.data_source = None
        self.raster = None

    def test_integrate_idw(self):
        hspatial.integrate(
            self.dataset, self.data_layer, self.target_band, hspatial.idw
        )
        result = self.target_band.ReadAsArray()
        nodatavalue = self.target_band.GetNoDataValue()

        # All masked points and only those must be no data
        # (^ is bitwise xor in Python)
        self.assertTrue(((result == nodatavalue) ^ (self.mask != 0)).all())

        self.assertAlmostEqual(result[3, 3], 62.971, places=3)
        self.assertAlmostEqual(result[6, 14], 34.838, places=3)
        self.assertAlmostEqual(result[4, 13], 30.737, places=3)


class CreateOgrLayerFromTimeseriesTestCase(TestCase):
    def setUp(self):
        self.tempdir = tempfile.mkdtemp()

        # Create two time series
        with open(os.path.join(self.tempdir, "ts1"), "w") as f:
            f.write(
                textwrap.dedent(
                    """\
                                    Location=23.78743 37.97385 4326

                                    """
                )
            )
        with open(os.path.join(self.tempdir, "ts2"), "w") as f:
            f.write(
                textwrap.dedent(
                    """\
                                    Location=24.56789 38.76543 4326

                                    """
                )
            )

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    def test_create_ogr_layer_from_timeseries(self):
        data_source = ogr.GetDriverByName("memory").CreateDataSource("tmp")
        filenames = [os.path.join(self.tempdir, x) for x in ("ts1", "ts2")]
        layer = hspatial.create_ogr_layer_from_timeseries(filenames, 2100, data_source)
        self.assertTrue(layer.GetFeatureCount(), 2)
        # The co-ordinates below are converted from WGS84 to Greek Grid/GGRS87
        ref = [
            {
                "x": 481180.63,
                "y": 4202647.01,  # 23.78743, 37.97385
                "filename": filenames[0],
            },
            {
                "x": 549187.44,
                "y": 4290612.25,  # 24.56789, 38.76543
                "filename": filenames[1],
            },
        ]
        for feature, r in zip(layer, ref):
            self.assertEqual(feature.GetField("filename"), r["filename"])
            self.assertAlmostEqual(feature.GetGeometryRef().GetX(), r["x"], 2)
            self.assertAlmostEqual(feature.GetGeometryRef().GetY(), r["y"], 2)


class HIntegrateTestCase(TestCase):
    """
    We will test on a 4x3 raster:

                        ABCD
                       1
                       2
                       3

    There will be three stations: two will be exactly at the center of
    gridpoints A3 and B3, and one will be slightly outside the grid (somewhere
    in E2). We assume the bottom left corner of A3 to have co-ordinates (0, 0)
    and the gridpoint to be 10 km across. We also consider C1 to be masked out.

    The calculations for the test have been made manually in
    data/spatial_calculations.ods, tab test_h_integrate.
    """

    def create_mask(self):
        mask_array = np.ones((3, 4), np.int8)
        mask_array[0, 2] = 0
        self.mask = gdal.GetDriverByName("mem").Create("mask", 4, 3, 1, gdal.GDT_Byte)
        self.mask.SetGeoTransform((0, 10000, 0, 30000, 0, -10000))
        self.mask.GetRasterBand(1).WriteArray(mask_array)

    def setUp(self):
        self.tempdir = tempfile.mkdtemp()

        self.filenames = [
            os.path.join(self.tempdir, x) for x in ("ts1", "ts2", "ts3", "ts4")
        ]
        with open(self.filenames[0], "w") as f:
            f.write(
                "Unit=microchips\n"
                "Timezone=EET (UTC+0200)\n"
                "Location=19.557285 0.0473312 2100\n"  # GGRS87=5000 5000
                "\n"
                "2014-04-22 12:50,5.03,\n"
                "2014-04-22 13:00,0.54,\n"
                "2014-04-22 13:10,6.93,\n"
            )
        with open(self.filenames[1], "w") as f:
            f.write(
                "Unit=microchips\n"
                "Timezone=EET (UTC+0200)\n"
                "Location=19.64689 0.04734 2100\n"  # GGRS87=15000 5000
                "\n"
                "2014-04-22 12:50,2.90,\n"
                "2014-04-22 13:00,2.40,\n"
                "2014-04-22 13:10,9.16,\n"
            )
        with open(self.filenames[2], "w") as f:
            f.write(
                "Unit=microchips\n"
                "Timezone=EET (UTC+0200)\n"
                "Location=19.88886 0.12857 2100\n"  # GGRS87=42000 14000
                "\n"
                "2014-04-22 12:50,9.70,\n"
                "2014-04-22 13:00,1.84,\n"
                "2014-04-22 13:10,7.63,\n"
            )
        with open(self.filenames[3], "w") as f:
            # This station is missing the date required,
            # so it should not be taken into account
            f.write(
                "Unit=microchips\n"
                "Timezone=EET (UTC+0200)\n"
                "Location=19.66480 0.15560 2100\n"  # GGRS87=17000 17000
                "\n"
                "2014-04-22 12:50,9.70,\n"
                "2014-04-22 13:10,7.63,\n"
            )

        self.create_mask()
        self.stations = ogr.GetDriverByName("memory").CreateDataSource("tmp")
        self.stations_layer = hspatial.create_ogr_layer_from_timeseries(
            self.filenames, 2100, self.stations
        )

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    def test_h_integrate(self):
        output_filename_prefix = os.path.join(self.tempdir, "test")
        result_filename = output_filename_prefix + "-2014-04-22-13-00+0200.tif"
        hspatial.h_integrate(
            mask=self.mask,
            stations_layer=self.stations_layer,
            date=dt.datetime(
                2014, 4, 22, 13, 0, tzinfo=TzinfoFromString("EET (+0200)")
            ),
            output_filename_prefix=output_filename_prefix,
            date_fmt="%Y-%m-%d %H:%M%z",
            funct=hspatial.idw,
            kwargs={},
        )
        f = gdal.Open(result_filename)
        result = f.GetRasterBand(1).ReadAsArray()
        nodatavalue = f.GetRasterBand(1).GetNoDataValue()
        expected_result = np.array(
            [
                [1.5088, 1.6064, nodatavalue, 1.7237],
                [1.3828, 1.6671, 1.7336, 1.7662],
                [0.5400, 2.4000, 1.7954, 1.7504],
            ]
        )
        np.testing.assert_almost_equal(result, expected_result, decimal=4)
        self.assertEqual(f.GetMetadataItem("UNIT"), "microchips")
        f = None

        # Wait long enough to make sure that, if we write to a file, its
        # modification time will be distinguishable from the modification time
        # of any file that has been written to so far (how long we need to wait
        # depends on the file system, so we use a test file).
        mtime_test_file = os.path.join(self.tempdir, "test_mtime")
        with open(mtime_test_file, "w") as f:
            f.write("hello, world")
        reference_mtime = os.path.getmtime(mtime_test_file)
        while os.path.getmtime(mtime_test_file) - reference_mtime < 0.001:
            sleep(0.001)
            with open(mtime_test_file, "w") as f:
                f.write("hello, world")

        # Try re-calculating the output; the output file should not be touched
        # at all.
        result_mtime = os.path.getmtime(result_filename)
        hspatial.h_integrate(
            mask=self.mask,
            stations_layer=self.stations_layer,
            date=dt.datetime(
                2014, 4, 22, 13, 0, tzinfo=TzinfoFromString("EET (+0200)")
            ),
            output_filename_prefix=output_filename_prefix,
            date_fmt="%Y-%m-%d %H:%M%z",
            funct=hspatial.idw,
            kwargs={},
        )
        self.assertEqual(os.path.getmtime(result_filename), result_mtime)

        # Now change one of the input files so that it contains new data that
        # can be used in the same calculation, and try recalculating. This time
        # the file should be recalculated.
        with open(self.filenames[3], "w") as f:
            f.write(
                "Timezone=EET (UTC+0200)\n"
                "Location=19.66480 0.15560 2100\n"  # GGRS87=17000 17000
                "\n"
                "2014-04-22 12:50,9.70,\n"
                "2014-04-22 13:00,4.70,\n"
                "2014-04-22 13:10,7.63,\n"
            )
        hspatial.h_integrate(
            mask=self.mask,
            stations_layer=self.stations_layer,
            date=dt.datetime(
                2014, 4, 22, 13, 0, tzinfo=TzinfoFromString("EET (+0200)")
            ),
            output_filename_prefix=output_filename_prefix,
            date_fmt="%Y-%m-%d %H:%M%z",
            funct=hspatial.idw,
            kwargs={},
        )
        self.assertGreater(os.path.getmtime(result_filename), result_mtime + 0.0009)
        f = gdal.Open(result_filename)
        result = f.GetRasterBand(1).ReadAsArray()
        nodatavalue = f.GetRasterBand(1).GetNoDataValue()
        expected_result = np.array(
            [
                [2.6736, 3.1053, nodatavalue, 2.5166],
                [2.3569, 3.5775, 2.9512, 2.3596],
                [0.5400, 2.4000, 2.5377, 2.3779],
            ]
        )
        np.testing.assert_almost_equal(result, expected_result, decimal=4)
        f = None


class ExtractPointFromRasterTestCase(TestCase):
    def setUp(self):
        self.tempdir = tempfile.mkdtemp()
        self._setup_test_raster()
        self.fp = gdal.Open(self.filename)

    def _setup_test_raster(self):
        self.filename = os.path.join(self.tempdir, "test_raster")
        nan = float("nan")
        setup_test_raster(
            self.filename,
            np.array([[1.1, nan, 1.3], [2.1, 2.2, nan], [3.1, 3.2, 3.3]]),
            dt.datetime(2014, 11, 21, 16, 1),
        )

    def tearDown(self):
        self.fp = None
        shutil.rmtree(self.tempdir)

    def test_top_left_point(self):
        point = hspatial.coordinates2point(22.005, 37.995)
        self.assertAlmostEqual(
            hspatial.extract_point_from_raster(point, self.fp), 1.1, places=2
        )

    def test_top_left_point_as_geodjango(self):
        point = GeoDjangoPoint(22.005, 37.995)
        self.assertAlmostEqual(
            hspatial.extract_point_from_raster(point, self.fp), 1.1, places=2
        )

    def test_top_middle_point(self):
        point = hspatial.coordinates2point(22.015, 37.995)
        self.assertTrue(math.isnan(hspatial.extract_point_from_raster(point, self.fp)))

    def test_middle_point(self):
        # We use co-ordinates almost to the common corner of the four lower left points,
        # only a little bit towards the center.
        point = hspatial.coordinates2point(22.01001, 37.98001)
        self.assertAlmostEqual(
            hspatial.extract_point_from_raster(point, self.fp), 2.2, places=2
        )

    def test_middle_point_with_GDALRaster(self):
        # Same as test_middle_point(), but uses GDALRaster object instead of a gdal
        # data source.
        point = hspatial.coordinates2point(22.01001, 37.98001)
        gdal_raster_object = GDALRaster(self.filename)
        self.assertAlmostEqual(
            hspatial.extract_point_from_raster(point, gdal_raster_object), 2.2, places=2
        )

    def test_bottom_left_point(self):
        # Use almost exactly same point as test_middle_point(), only slightly altered
        # so that we get bottom left point instead.
        point = hspatial.coordinates2point(22.00999, 37.97999)
        self.assertAlmostEqual(
            hspatial.extract_point_from_raster(point, self.fp), 3.1, places=2
        )

    def test_middle_point_with_GRS80(self):
        # Same as test_middle_point(), but with a different reference system, GRS80; the
        # result should be the same.
        point = hspatial.coordinates2point(325077, 4205177, srid=2100)
        self.assertAlmostEqual(
            hspatial.extract_point_from_raster(point, self.fp), 2.2, places=2
        )

    def test_does_not_modify_srid_of_point(self):
        point = hspatial.coordinates2point(325077, 4205177, srid=2100)
        original_spatial_reference = point.GetSpatialReference().ExportToWkt()
        hspatial.extract_point_from_raster(point, self.fp)
        self.assertEqual(
            point.GetSpatialReference().ExportToWkt(), original_spatial_reference
        )

    def test_bottom_left_point_with_GRS80(self):
        # Same as test_bottom_left_point(), but with a different reference system,
        # GRS80; the result should be the same.
        point = hspatial.coordinates2point(324076, 4205176, srid=2100)
        self.assertAlmostEqual(
            hspatial.extract_point_from_raster(point, self.fp), 3.1, places=2
        )

    def test_point_outside_raster(self):
        point = hspatial.coordinates2point(21.0, 38.0)
        with self.assertRaises(RuntimeError):
            hspatial.extract_point_from_raster(point, self.fp)


class ExtractPointFromRasterWhenOutsideCRSLimitsTestCase(TestCase):
    def setUp(self):
        self.tempdir = tempfile.mkdtemp()
        self._setup_test_raster()
        self.fp = gdal.Open(self.filename)

    def _setup_test_raster(self):
        self.filename = os.path.join(self.tempdir, "test_raster")
        nan = float("nan")
        setup_test_raster(
            self.filename,
            np.array([[1.1, nan, 1.3], [2.1, 2.2, nan], [3.1, 3.2, 3.3]]),
            dt.datetime(2014, 11, 21, 16, 1),
            srid=2100,
        )

    def tearDown(self):
        self.fp = None
        shutil.rmtree(self.tempdir)

    def test_fails_gracefully_when_osr_point_is_really_outside_crs_limits(self):
        point = hspatial.coordinates2point(125.0, 85.0)
        with self.assertRaises(RuntimeError):
            hspatial.extract_point_from_raster(point, self.fp)

    def test_fails_gracefully_when_geos_point_is_really_outside_crs_limits(self):
        point = GeoDjangoPoint(125.0, 85.0)
        with self.assertRaises(RuntimeError):
            hspatial.extract_point_from_raster(point, self.fp)


class SetupTestRastersMixin:
    include_time = True

    def setUp(self):
        self.tempdir = tempfile.mkdtemp()
        self._setup_test_rasters()

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    def _setup_test_rasters(self):
        self._setup_raster(
            dt.date(2014, 11, 21),
            np.array([[1.1, 1.2, 1.3], [2.1, 2.2, 2.3], [3.1, 3.2, 3.3]]),
        )
        self._setup_raster(
            dt.date(2014, 11, 22),
            np.array([[11.1, 12.1, 13.1], [21.1, 22.1, 23.1], [31.1, 32.1, 33.1]]),
        )
        self._setup_raster(
            dt.date(2014, 11, 23),
            np.array(
                [[110.1, 120.1, 130.1], [210.1, 220.1, 230.1], [310.1, 320.1, 330.1]]
            ),
        )

    def _setup_raster(self, date, value):
        filename = self._create_filename(date)
        timestamp = self._create_timestamp(date)
        setup_test_raster(filename, value, timestamp, unit="microkernels")

    def _create_filename(self, date):
        result = date.strftime("test-%Y-%m-%d")
        if self.include_time:
            result += "-16-1"
        result += ".tif"
        return os.path.join(self.tempdir, result)

    def _create_timestamp(self, date):
        if self.include_time:
            return dt.datetime.combine(date, dt.time(16, 1, tzinfo=UTC_PLUS_2))
        else:
            return date

    def _check_against_expected(self, ts):
        expected = pd.DataFrame(
            data={"value": [2.2, 22.1, 220.1], "flags": ["", "", ""]},
            index=self.expected_index,
            columns=["value", "flags"],
        )
        expected.index.name = "date"
        self.assertEqual(
            ts.data.index.tz.utcoffset(None), expected.index.tz.utcoffset(None)
        )
        pd.testing.assert_frame_equal(ts.data, expected, check_index_type=False)

    @property
    def expected_index(self):
        hour, minute = self.include_time and (16, 1) or (23, 58)
        return [
            dt.datetime(2014, 11, 21, hour, minute, tzinfo=UTC_PLUS_2),
            dt.datetime(2014, 11, 22, hour, minute, tzinfo=UTC_PLUS_2),
            dt.datetime(2014, 11, 23, hour, minute, tzinfo=UTC_PLUS_2),
        ]


class PointTimeseriesGetTestCase(SetupTestRastersMixin, TestCase):
    def test_with_list_of_files(self):
        # Use co-ordinates almost to the common corner of the four lower left points,
        # and only a little bit towards the center.
        point = hspatial.coordinates2point(22.01001, 37.98001)
        filenames = [
            os.path.join(self.tempdir, "test-2014-11-22-16-1.tif"),
            os.path.join(self.tempdir, "test-2014-11-21-16-1.tif"),
            os.path.join(self.tempdir, "test-2014-11-23-16-1.tif"),
        ]
        ts = hspatial.PointTimeseries(
            point, filenames=filenames, default_time=dt.time(0, 0, tzinfo=UTC_PLUS_2)
        ).get()
        self._check_against_expected(ts)

    def test_raises_when_no_timezone(self):
        point = hspatial.coordinates2point(22.01001, 37.98001)
        filenames = [os.path.join(self.tempdir, "test-2014-11-22-16-1.tif")]
        with self.assertRaises(TypeError):
            hspatial.PointTimeseries(
                point, filenames=filenames, default_time=dt.time(0, 0)
            )

    def test_with_prefix(self):
        # Same as test_with_list_of_files(), but with prefix.
        point = hspatial.coordinates2point(22.01001, 37.98001)
        prefix = os.path.join(self.tempdir, "test")
        ts = hspatial.PointTimeseries(
            point, prefix=prefix, default_time=dt.time(0, 0, tzinfo=UTC_PLUS_2)
        ).get()
        self._check_against_expected(ts)

    def test_with_prefix_and_geodjango(self):
        point = hspatial.coordinates2point(22.01001, 37.98001)
        prefix = os.path.join(self.tempdir, "test")
        ts = hspatial.PointTimeseries(
            point, prefix=prefix, default_time=dt.time(0, 0, tzinfo=UTC_PLUS_2)
        ).get()
        self._check_against_expected(ts)

    def test_unit_of_measurement(self):
        point = hspatial.coordinates2point(22.01001, 37.98001)
        prefix = os.path.join(self.tempdir, "test")
        ts = hspatial.PointTimeseries(
            point, prefix=prefix, default_time=dt.time(0, 0, tzinfo=UTC_PLUS_2)
        ).get()
        self.assertEqual(ts.unit, "microkernels")


class PointTimeseriesGetDailyTestCase(SetupTestRastersMixin, TestCase):
    include_time = False

    def test_with_list_of_files(self):
        # Use co-ordinates almost to the center of the four lower left points, and only
        # a little bit towards the center.
        point = hspatial.coordinates2point(22.01001, 37.98001)
        filenames = [
            os.path.join(self.tempdir, "test-2014-11-22.tif"),
            os.path.join(self.tempdir, "test-2014-11-21.tif"),
            os.path.join(self.tempdir, "test-2014-11-23.tif"),
        ]
        ts = hspatial.PointTimeseries(
            point, filenames=filenames, default_time=dt.time(23, 58, tzinfo=UTC_PLUS_2)
        ).get()
        self._check_against_expected(ts)

    def test_with_prefix(self):
        # Same as test_with_list_of_files(), but with prefix.
        point = hspatial.coordinates2point(22.01001, 37.98001)
        prefix = os.path.join(self.tempdir, "test")
        ts = hspatial.PointTimeseries(
            point, prefix=prefix, default_time=dt.time(23, 58, tzinfo=UTC_PLUS_2)
        ).get()
        self._check_against_expected(ts)

    def test_with_prefix_and_geodjango(self):
        point = GeoDjangoPoint(22.01001, 37.98001)
        prefix = os.path.join(self.tempdir, "test")
        ts = hspatial.PointTimeseries(
            point, prefix=prefix, default_time=dt.time(23, 58, tzinfo=UTC_PLUS_2)
        ).get()
        self._check_against_expected(ts)


class PointTimeseriesGetCachedTestCase(SetupTestRastersMixin, TestCase):
    def setUp(self):
        super().setUp()
        self.point = hspatial.coordinates2point(22.01001, 37.98001)
        self.prefix = os.path.join(self.tempdir, "test")
        self.dest = os.path.join(self.tempdir, "dest.hts")

    def test_result(self):
        result = hspatial.PointTimeseries(
            self.point,
            prefix=self.prefix,
            default_time=dt.time(0, 0, tzinfo=UTC_PLUS_2),
        ).get_cached(self.dest)
        self._check_against_expected(result)

    def test_file(self):
        hspatial.PointTimeseries(
            self.point,
            prefix=self.prefix,
            default_time=dt.time(0, 0, tzinfo=UTC_PLUS_2),
        ).get_cached(self.dest)
        with open(self.dest, "r", newline="\n") as f:
            self._check_against_expected(HTimeseries(f, default_tzinfo=UTC_PLUS_2))

    def test_version(self):
        hspatial.PointTimeseries(
            self.point,
            prefix=self.prefix,
            default_time=dt.time(0, 0, tzinfo=UTC_PLUS_2),
        ).get_cached(self.dest, version=2)
        with open(self.dest, "r") as f:
            first_line = f.readline()
        self.assertEqual(first_line, "Version=2\n")

    def test_file_is_not_recreated(self):
        hspatial.PointTimeseries(
            self.point,
            prefix=self.prefix,
            default_time=dt.time(0, 0, tzinfo=UTC_PLUS_2),
        ).get_cached(self.dest)

        # Make existing file read-only
        os.chmod(self.dest, S_IREAD | S_IRGRP | S_IROTH)

        # Try again—it shouldn't try to write, therefore it shouldn't raise exception
        hspatial.PointTimeseries(
            self.point,
            prefix=self.prefix,
            default_time=dt.time(0, 0, tzinfo=UTC_PLUS_2),
        ).get_cached(self.dest)
        with open(self.dest, "r", newline="\n") as f:
            self._check_against_expected(HTimeseries(f, default_tzinfo=UTC_PLUS_2))

    def test_file_is_recreated_when_out_of_date(self):
        hspatial.PointTimeseries(
            self.point,
            prefix=self.prefix,
            default_time=dt.time(0, 0, tzinfo=UTC_PLUS_2),
        ).get_cached(self.dest)
        self._setup_additional_raster()

        # Make existing file read-only
        os.chmod(self.dest, S_IREAD | S_IRGRP | S_IROTH)

        # Try again—it should raise exception
        with self.assertRaises(PermissionError):
            hspatial.PointTimeseries(
                self.point,
                prefix=self.prefix,
                default_time=dt.time(0, 0, tzinfo=UTC_PLUS_2),
            ).get_cached(self.dest)

    def _setup_additional_raster(self):
        filename = os.path.join(self.tempdir, "test-2014-11-24-16-1.tif")
        setup_test_raster(
            filename,
            np.array(
                [[110.1, 120.1, 130.1], [210.1, 220.1, 230.1], [310.1, 320.1, 330.1]]
            ),
            dt.datetime(2014, 11, 24, 16, 1),
        )

    def test_start_date(self):
        start_date = dt.datetime(2014, 11, 22, 16, 1)
        result = hspatial.PointTimeseries(
            self.point,
            prefix=self.prefix,
            start_date=start_date,
            default_time=dt.time(0, 0, tzinfo=UTC_PLUS_2),
        ).get_cached(self.dest)
        self.assertEqual(result.data.index[0], start_date.replace(tzinfo=UTC_PLUS_2))

    def test_end_date(self):
        end_date = dt.datetime(2014, 11, 22, 16, 1)
        result = hspatial.PointTimeseries(
            self.point,
            prefix=self.prefix,
            end_date=end_date,
            default_time=dt.time(0, 0, tzinfo=UTC_PLUS_2),
        ).get_cached(self.dest)
        self.assertEqual(result.data.index[-1], end_date.replace(tzinfo=UTC_PLUS_2))


class FilenameWithDateFormatTestCase(TestCase):
    def test_with_given_datetime_format(self):
        format = hspatial.FilenameWithDateFormat(
            "myprefix", date_fmt="%d-%m-%Y-%H-%M", tzinfo=UTC_PLUS_2
        )
        self.assertEqual(
            format.get_date("myprefix-4-8-2019-10-41.tif"),
            dt.datetime(2019, 8, 4, 10, 41, tzinfo=UTC_PLUS_2),
        )

    def test_with_given_date_format(self):
        format = hspatial.FilenameWithDateFormat(
            "myprefix", date_fmt="%d-%m-%Y", tzinfo=UTC_PLUS_2
        )
        self.assertEqual(
            format.get_date("myprefix-4-8-2019.tif"),
            dt.datetime(2019, 8, 4, tzinfo=UTC_PLUS_2),
        )

    def test_datetime_with_auto_format(self):
        format = hspatial.FilenameWithDateFormat("myprefix", tzinfo=UTC_PLUS_2)
        self.assertEqual(
            format.get_date("myprefix-2019-8-4-10-41.tif"),
            dt.datetime(2019, 8, 4, 10, 41, tzinfo=UTC_PLUS_2),
        )

    def test_date_with_auto_format(self):
        format = hspatial.FilenameWithDateFormat("myprefix", tzinfo=UTC_PLUS_2)
        self.assertEqual(
            format.get_date("myprefix-2019-8-4.tif"),
            dt.datetime(2019, 8, 4, tzinfo=UTC_PLUS_2),
        )


class PassepartoutPointTestCase(TestCase):
    def test_transform_does_not_modify_srid_of_gdal_point(self):
        pppoint = hspatial.PassepartoutPoint(
            hspatial.coordinates2point(324651, 4205742, srid=2100)
        )
        original_spatial_reference = pppoint.point.GetSpatialReference().ExportToWkt()
        sr = osr.SpatialReference()
        sr.ImportFromEPSG(4326)
        pppoint.transform_to(sr.ExportToWkt())
        self.assertEqual(
            pppoint.point.GetSpatialReference().ExportToWkt(),
            original_spatial_reference,
        )

    def test_transform_does_not_modify_srid_of_geodjango_point(self):
        pppoint = hspatial.PassepartoutPoint(GeoDjangoPoint(324651, 4205742, srid=2100))
        pppoint.transform_to(4326)
        self.assertEqual(pppoint.point.srid, 2100)



# cython: language_level=3
import datetime as dt

cimport numpy as np
from cpython cimport array

import numpy as np
import pandas as pd


class Rocc:
    def __init__(self, timeseries, thresholds, symmetric, flag):
        self.htimeseries = timeseries
        self.thresholds = thresholds
        self.symmetric = symmetric
        self.flag = flag or ""

    def execute(self):
        self._transform_thresholds()
        self._transform_to_plain_numpy()
        failures = self._do_actual_job()
        self._transform_to_pandas()
        return failures

    def _transform_thresholds(self):
        threshold_deltas = array.array("l")
        threshold_allowed_diffs = array.array("d")

        for threshold in self.thresholds:
            delta_t = int(self._get_delta_t_transformed(threshold.delta_t))
            threshold_deltas.append(delta_t)
            threshold_allowed_diffs.append(threshold.allowed_diff)
        self.threshold_deltas = threshold_deltas
        self.threshold_allowed_diffs = threshold_allowed_diffs

    def _get_delta_t_transformed(self, delta_t):
        if not delta_t[0].isdigit():
            delta_t = "1" + delta_t
        return pd.Timedelta(delta_t).to_timedelta64()

    def _transform_to_plain_numpy(self):
        flag_lengths = self.htimeseries.data["flags"].str.len()
        max_flag_length = 0 if flag_lengths.empty else max(flag_lengths)
        flags_dtype = "U" + str(max_flag_length + 1 + len(self.flag))
        self.ts_index = self.htimeseries.data.index.values.astype(long)
        self.ts_values = self.htimeseries.data["value"].values
        self.ts_flags = self.htimeseries.data["flags"].values.astype(flags_dtype)
        try:
            utc_offset = self.htimeseries.data.index.tz.utcoffset(dt.datetime.now())
        except AttributeError:
            utc_offset = dt.timedelta(0)
        self.ts_utc_offset_minutes = int(utc_offset.total_seconds() / 60)

    def _do_actual_job(self):
        return _perform_rocc(
            self.ts_index,
            self.ts_values,
            self.ts_flags,
            self.ts_utc_offset_minutes,
            list(self.thresholds),
            self.threshold_deltas,
            self.threshold_allowed_diffs,
            self.symmetric,
            self.flag,
        )

    def _transform_to_pandas(self):
        self.htimeseries.data = pd.DataFrame(
            index=self.htimeseries.data.index,
            columns=["value", "flags"],
            data=np.vstack((self.ts_values, self.ts_flags)).transpose(),
        )
        self.htimeseries.data["value"] = self.htimeseries.data["value"].astype(np.float64)


# IMPORTANT: There's some plain Python in the Cython below. Specifically, there are some
# Python lists and some places with undeclared variables. These are only used when a
# failure is found. Given that failures should be very few, this should not affect the
# overall speed. But I'm not really a Cython expert and I don't know exactly how it
# works.


def _perform_rocc(
    np.ndarray ts_index,
    np.ndarray ts_values,
    np.ndarray ts_flags,
    int ts_utc_offset_minutes,
    list thresholds,
    array.array threshold_deltas,
    array.array threshold_allowed_diffs,
    int symmetric,
    str flag,
):
    cdef int i, record_fails_check
    cdef list failures = []

    for i in range(ts_index.size):
        record_fails_check = _record_fails_check(
            i,
            ts_index,
            ts_values,
            ts_utc_offset_minutes,
            thresholds,
            threshold_deltas,
            threshold_allowed_diffs,
            symmetric,
            failures,
        )
        if record_fails_check and flag:
            _add_flag(i, ts_flags, flag)
    return failures


def _add_flag(int i, np.ndarray ts_flags, str flag):
    if ts_flags[i]:
        ts_flags[i] = ts_flags[i] + " "
    ts_flags[i] = ts_flags[i] + flag


def _record_fails_check(
    int record_index,
    np.ndarray ts_index,
    np.ndarray ts_values,
    int ts_utc_offset_minutes,
    list thresholds,
    array.array threshold_deltas,
    array.array threshold_allowed_diffs,
    int symmetric,
    list failures,
):
    cdef int ti
    cdef double diff

    for ti in range(len(threshold_deltas)):
        diff = _record_fails_threshold(
            record_index,
            threshold_deltas[ti],
            threshold_allowed_diffs[ti],
            ts_index,
            ts_values,
            symmetric,
        )
        if diff:
            timestamp = ts_index[record_index].item()
            datestr = str(
                np.datetime64(timestamp, "ns") + np.timedelta64(ts_utc_offset_minutes, "m")
            )[:16]
            diffsign = '+' if diff > 0 else ''
            thresholdsign = '-' if diff < 0 else ''
            cmpsign = '>' if diff > 0 else '<'
            failures.append(
                f"{datestr}  {diffsign}{diff} in {thresholds[ti].delta_t} "
                f"({cmpsign} {thresholdsign}{threshold_allowed_diffs[ti]})"
            )
            return True
    return False


def _record_fails_threshold(
    int record_index,
    long threshold_delta,
    double threshold_allowed_diff,
    np.ndarray ts_index,
    np.ndarray ts_values,
    int symmetric,
):
    cdef double current_value = ts_values[record_index]
    cdef long current_timestamp = ts_index[record_index]
    cdef int i, fails
    cdef double diff;

    for i in range(record_index - 1, -1, -1):
        if current_timestamp - ts_index[i] > threshold_delta:
            return False
        diff = current_value - ts_values[i];
        fails = (
            diff > threshold_allowed_diff
            or (symmetric and diff < -threshold_allowed_diff)
        )
        if fails:
            return diff
    return False
